# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)


import tempfile
import shutil
import os

class MakeTempDir(object):
    temp_dir = None

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()
        logger.debug('Using temporary directory %s...', self.temp_dir)
        return os.path.realpath(self.temp_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            logger.debug('Cleaning up temporary directory %s...', self.temp_dir)
            shutil.rmtree(self.temp_dir)
        except Exception, e:
            logger.exception('Failure cleaning up temp space')
            pass



#!/usr/bin/env python

import subprocess
import argparse
import sys
from subprocess import CalledProcessError, STDOUT
from time import sleep


def wait_on_hosts(hosts, max_attempts=3, sleep_time=1):
    '''
    Wait for a container to have a State.Running value = true.
    :param hosts: list of service names taken from container.yml
    :param max_attempts: Max number of times to inspect the container and check State.Running
    :param sleep_time: Number of seconds to wait between attempts.
    :return: dict of host:running pairs
    '''
    results = {}
    for host in hosts:
        container = "ansible_{}_1".format(host)
        tries = max_attempts
        host_ready = False
        output = None
        results[host] = False
        while tries > 0 and not host_ready:
            try:
                output = subprocess.check_output(["docker", "inspect", "--format", "{{ .State.Running }}",
                                                  container], stderr=STDOUT)
            except CalledProcessError:
                pass
            tries -= 1
            if output and 'true' in output:
                host_ready = True
                results[host] = True
            else:
                sleep(sleep_time)
    return results

if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='wait_on_host',
                                     description='Wait for a host or list of hosts to be in a running state')
    parser.add_argument('--max-attempts', '-m', type=int, action='store', default=3,
                        help=u"max number of attempts at checking a container's state, defaults to 3")
    parser.add_argument('--sleep-time', '-s', type=int, action='store', default=1,
                        help=u'number of seconds to wait between attempts, defaults to 1')
    parser.add_argument('host', nargs='+',
                        help=u'name of the host to wait on')
    args = parser.parse_args()

    if args.host:
        results = wait_on_hosts(args.host, max_attempts=args.max_attempts, sleep_time=args.sleep_time)
        status = 0
        for host, running in results.iteritems():
            print "Host {0} {1}".format(host, 'running' if running else 'failed')
            if not running:
                status = 1
        sys.exit(status)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import os
import sys
import argparse

from . import engine
from . import exceptions
from .utils import load_shipit_engine, AVAILABLE_SHIPIT_ENGINES

from logging import config
LOGGING = {
        'version': 1,
        'disable_existing_loggers': True,
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
            },
        },
        'loggers': {
            'container': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False
            },
            'compose': {
                'handlers': ['console'],
                'level': 'WARNING',
                'propagate': False
            },
            'docker': {
                'handlers': ['console'],
                'level': 'WARNING',
                'propagate': False
            }
        },
        'root': {
            'handlers': ['console'],
            'level': 'ERROR'
        }
    }


AVAILABLE_COMMANDS = {'help': 'Display this help message',
                      'version': 'Display Ansible Container version information',
                      'init': 'Initialize a new Ansible Container project',
                      'install': 'Install a service from Ansible Galaxy',
                      'build': 'Build new images based on ansible/container.yml',
                      'run': 'Run and orchestrate built images based on container.yml',
                      'stop': 'Stop the services defined in container.yml, if deployed',
                      'restart': 'Restart the services defined in container.yml',
                      'push': 'Push your built images to a Docker Hub compatible registry',
                      'shipit': 'Generate a deployment playbook to your cloud of choice.'}

def subcmd_common_parsers(parser, subparser, cmd):
    if cmd in ('build', 'run', 'shipit', 'push'):
        subparser.add_argument('--with-volumes', '-v', action='store', nargs='+',
                               help=u'Mount one or more volumes to the Ansible Builder Container. '
                                    u'Specify volumes as strings using the Docker volume format.',
                               default=[])
        subparser.add_argument('--with-variables', '-e', action='store', nargs='+',
                               help=u'Define one or more environment variables in the Ansible '
                                    u'Builder Container. Format each variable as a key=value string.',
                               default=[])
        subparser.add_argument('--roles-path', action='store', default=None,
                               help=u'Specify a local path containing roles you want to '
                                    u'use in the builder container.')

def subcmd_init_parser(parser, subparser):
    subparser.add_argument('--server', '-s', action='store',
                           default='https://galaxy.ansible.com/',
                           help=u'Use a different Galaxy server URL')
    subparser.add_argument('project', nargs='?', action='store',
                           help=u'Use a project template instead of making a '
                                u'blank project from an Ansible Container project '
                                u'from Ansible Galaxy.')

def subcmd_build_parser(parser, subparser):
    subparser.add_argument('--flatten', action='store_true',
                           help=u'By default, Ansible Container will add a single '
                                u'layer to your base images. Specify this to squash '
                                u'the images down to a single layer.',
                           dest='flatten', default=False)
    subparser.add_argument('--no-purge-last', action='store_false',
                           help=u'By default, Ansible Container will remove the '
                                u'previously built image for your hosts. Disable '
                                u'that with this flag.')
    subparser.add_argument('--from-scratch', action='store_true',
                           help=u'Instead of running the Ansible playbook against '
                                u'the existing copies of your containers, run the '
                                u'playbook against the base image, rebuilding them '
                                u'from scratch.',
                           dest='rebuild', default=False)
    subparser.add_argument('--local-builder', action='store_true',
                           help=u'Instead of using the Ansible Builder Container '
                                u'image from Docker Hub, generate one locally.')
    subparser.add_argument('--save-build-container', action='store_true',
                           help=u'Leave the Ansible Builder Container intact upon build completion. '
                                u'Use for debugging and testing.', default=False)
    subparser.add_argument('--services', action='store',
                           help=u'Rather than perform an orchestrated build, only build specific services.',
                           nargs='+', dest='service', default=None)
    subparser.add_argument('ansible_options', action='store',
                           help=u'Provide additional commandline arguments to '
                                u'Ansible in executing your playbook. If you '
                                u'use this argument, you will need to use -- to '
                                u'prefix your extra options. Use this feature with '
                                u'caution.', default=u'', nargs='*')
    subcmd_common_parsers(parser, subparser, 'build')

def subcmd_run_parser(parser, subparser):
    subparser.add_argument('service', action='store',
                           help=u'The specific services you want to run',
                           nargs='*')
    subparser.add_argument('--production', action='store_true',
                           help=u'Run the production configuration locally',
                           default=False, dest='production')
    subparser.add_argument('-d', '--detached', action='store_true',
                           help=u'Run the application in detached mode', dest='detached')
    subparser.add_argument('-o', '--remove-orphans', action='store_true',
                           help=u'Remove containers for services not defined in container.yml',
                           default=False, dest='remove_orphans')
    subcmd_common_parsers(parser, subparser, 'run')

def subcmd_stop_parser(parser, subparser):
    subparser.add_argument('service', action='store',
                           help=u'The specific services you want to stop',
                           nargs='*')
    subparser.add_argument('-f', '--force', action='store_true',
                           help=u'Force stop running containers',
                           dest='force')

def subcmd_restart_parser(parser, subparser):
    subparser.add_argument('service', action='store',
                           help=u'The specific services you want to restart',
                           nargs='*')

def subcmd_help_parser(parser, subparser):
    return

def subcmd_push_parser(parser, subparser):
    subparser.add_argument('--username', action='store',
                           help=u'If authentication with the registry is required, provide a valid username.',
                           dest='username', default=None)
    subparser.add_argument('--email', action='store',
                           help=(u'If authentication with the registry requires an email address, provide a '
                                 u'valid email address'),
                           dest='email', default=None)
    subparser.add_argument('--password', action='store',
                           help=u'If authentication with the registry is required, provide a valid password.',
                           dest='password', default=None)
    subparser.add_argument('--push-to', action='store',
                           help=(u'Name of a registry defined in container.yml or the actual URL of the registry, '
                                 u'including the namespace. If passing a URL, an example would be: '
                                 u'"https://registry.example.com:5000/myproject"'),
                           dest='push_to', default=None)
    subcmd_common_parsers(parser, subparser, 'push')

def subcmd_version_parser(parser, subparser):
    return

def subcmd_shipit_parser(parser, subparser):
    se_subparser = subparser.add_subparsers(title='shipit-engine', dest='shipit_engine')
    for engine_name, engine in AVAILABLE_SHIPIT_ENGINES.items():
        engine_parser = se_subparser.add_parser(engine_name, help=engine['help'])
        engine_obj = load_shipit_engine(engine['cls'], base_path=os.getcwd())
        engine_obj.add_options(engine_parser)
    subcmd_common_parsers(parser, subparser, 'shipit')

def subcmd_install_parser(parser, subparser):
    subparser.add_argument('roles', nargs='+', action='store')

def commandline():
    parser = argparse.ArgumentParser(description=u'Build, orchestrate, run, and '
                                                 u'ship Docker containers with '
                                                 u'Ansible playbooks')
    parser.add_argument('--debug', action='store_true', dest='debug',
                        help=u'Enable debug output', default=False)
    parser.add_argument('--engine', action='store', dest='engine_name',
                        help=u'Select your container engine and orchestrator',
                        default='docker')
    parser.add_argument('--project', '-p', action='store', dest='base_path',
                        help=u'Specify a path to your project. Defaults to '
                             u'current working directory.', default=os.getcwd())
    parser.add_argument('--var-file', action='store',
                        help=u'Path to a YAML or JSON formatted file providing variables for '
                             u'Jinja2 templating in container.yml.', default=None)

    subparsers = parser.add_subparsers(title='subcommand', dest='subcommand')
    for subcommand in AVAILABLE_COMMANDS:
        logger.debug('Registering subcommand %s', subcommand)
        subparser = subparsers.add_parser(subcommand, help=AVAILABLE_COMMANDS[subcommand])
        globals()['subcmd_%s_parser' % subcommand](parser, subparser)

    args = parser.parse_args()

    if args.subcommand == 'help':
        parser.print_help()
        sys.exit(0)

    if args.debug and args.subcommand != 'version':
        LOGGING['loggers']['container']['level'] = 'DEBUG'
    config.dictConfig(LOGGING)

    try:
        getattr(engine, u'cmdrun_{}'.format(args.subcommand))(**vars(args))
    except exceptions.AnsibleContainerAlreadyInitializedException, e:
        logger.error('Ansible Container is already initialized')
        sys.exit(1)
    except exceptions.AnsibleContainerNotInitializedException, e:
        logger.error('No Ansible Container project data found - do you need to '
                     'run "ansible-container init"?')
        sys.exit(1)
    except exceptions.AnsibleContainerNoAuthenticationProvidedException, e:
        logger.error(unicode(e))
        sys.exit(1)
    except exceptions.AnsibleContainerNoMatchingHosts:
        logger.error('No matching service found in ansible/container.yml')
        sys.exit(1)
    except exceptions.AnsibleContainerHostNotTouchedByPlaybook:
        logger.error('The requested service(s) is not referenced in ansible/main.yml. Nothing to build.')
        sys.exit(1)
    except Exception, e:
        if args.debug:
            logger.exception(unicode(e))
        else:
            logger.error(unicode(e))
        sys.exit(1)



#!/usr/bin/python
#
# Copyright 2016 Ansible by Red Hat
#
# This file is part of ansible-container
#

DOCUMENTATION = '''

module: k8s_route

short_description: Create or remove a route on a Kubernetes or OpenShift cluster.

description:
  - Create or remove a route on a Kubernetes or OpenShift cluster by setting the C(state) to I(present) or I(absent).
  - The module is idempotent and will not replace an existing route unless the C(reload) option is passed.
  - Supports check mode. Use check mode to view a list of actions the module will take.

options:

'''

EXAMPLES = '''
'''

RETURN = '''
'''
import logging
import logging.config

from ansible.module_utils.basic import *

logger = logging.getLogger('oso_route')

LOGGING = (
    {
        'version': 1,
        'disable_existing_loggers': True,
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
            },
            'file': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': 'ansible-container.log'
            }
        },
        'loggers': {
            'oso_route': {
                'handlers': ['file'],
                'level': 'DEBUG',
            },
            'container': {
                'handlers': ['file'],
                'level': 'DEBUG',
            },
            'compose': {
                'handlers': [],
                'level': 'INFO'
            },
            'docker': {
                'handlers': [],
                'level': 'INFO'
            }
        },
    }
)


class RouteManager(object):

    def __init__(self):

        self.arg_spec = dict(
            project_name=dict(type='str', aliases=['namespace']),
            state=dict(type='str', choices=['present', 'absent'], default='present'),
            labels=dict(type='dict'),
            route_name=dict(type='str'),
            host=dict(type='str'),
            service_name=dict(type='str', required=True, aliases=['to']),
            service_port=dict(type='str', required=True, aliases=['port']),
            replace=dict(type='bool', default=False),
            cli=dict(type='str', choices=['kubectl', 'oc'], default='oc'),
            debug=dict(type='bool', default=False)
        )

        self.module = AnsibleModule(self.arg_spec,
                                    supports_check_mode=True)

        self.project_name = None
        self.state = None
        self.labels = None
        self.route_name = None
        self.host = None
        self.service_name = None
        self.service_port = None
        self.replace = None
        self.cli = None
        self.api = None
        self.debug = None
        self.check_mode = self.module.check_mode

    def exec_module(self):

        for key in self.arg_spec:
            setattr(self, key, self.module.params.get(key))

        if self.debug:
            LOGGING['loggers']['container']['level'] = 'DEBUG'
            LOGGING['loggers']['oso_route']['level'] = 'DEBUG'
        logging.config.dictConfig(LOGGING)

        self.api = OriginAPI(self.module)

        actions = []
        changed = False
        routes = dict()
        results = dict()

        project_switch = self.api.set_project(self.project_name)
        if not project_switch:
            actions.append("Create project %s" % self.project_name)
            if not self.check_mode:
                self.api.create_project(self.project_name)

        if self.state == 'present':
            route = self.api.get_resource('route', self.route_name)
            if not route:
                template = self._create_template()
                changed = True
                actions.append("Create route %s" % self.route_name)
                if not self.check_mode:
                    self.api.create_from_template(template=template)
            elif route and self.replace:
                template = self._create_template()
                changed = True
                actions.append("Replace route %s" % self.route_name)
                if not self.check_mode:
                    self.api.replace_from_template(template=template)

            routes[self.route_name.replace('-', '_') + '_route'] = self.api.get_resource('route', self.route_name)

        elif self.state == 'absent':
            if self.api.get_resource('route', self.route_name):
                changed = True
                actions.append("Delete route %s" % self.route_name)
                if not self.check_mode:
                    self.api.delete_resource('route', self.route_name)

        results['changed'] = changed

        if self.check_mode:
            results['actions'] = actions

        if routes:
            results['ansible_facts'] = routes

        self.module.exit_json(**results)

    def _create_template(self):
        '''
        apiVersion: v1
        kind: Route
        metadata:
          name: wordpress-wordpress
          labels:
            wordpress: wordpress
        spec:
          host: wordpress.local
          to:
            kind: Service
            name: wordpress-wordpress
          port:
            targetPort: main
        '''

        template = dict(
            apiVersion="v1",
            kind="Route",
            metadata=dict(
                name=self.route_name,
            ),
            spec=dict(
                to=dict(
                    kind="Service",
                    name=self.service_name
                ),
                port=dict(
                    targetPort=self.service_port
                )
            )
        )

        if self.host:
            template['spec']['host'] = self.host

        if self.labels:
            template['metadata']['labels'] = self.labels

        return template

#The following will be included by `ansble-container shipit` when cloud modules are copied into the role library path.
#include--> oso_api.py


def main():
    manager = RouteManager()
    manager.exec_module()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
import logging

logger = logging.getLogger(__name__)


import re

from collections import OrderedDict


class Route(object):

    def __init__(self, config=None, project_name=None):
        self.project_name = project_name
        self.config = config

    def get_template(self):
        return self._get_task_or_config(request_type="config")

    def get_task(self):
        return self._get_task_or_config(request_type="task")

    def _get_task_or_config(self, request_type="task"):
        templates = []
        for name, service in self.config.get('services', {}).items():
            new_routes = self._create(request_type, name, service)
            if new_routes:
                templates += new_routes
        return templates

    def _create(self, type, name, service):
        '''
        Generate Openshift route templates or playbook tasks. Each port on a service definition
        represents an externally exposed port.
        '''

        templates = []
        hostname = None

        options = service.get('options', {}).get('openshift', {})
        state = options.get('state', 'present')

        if options.get('hostname'):
            hostname = options['hostname']

        service_ports = self._get_service_ports(service)

        if service_ports:
            for port in service_ports:
                route_name = "%s-%s" % (name, port)
                labels = dict(
                    app=self.project_name,
                    service=name
                )

                if type == 'config' and state != 'absent':
                    template = dict(
                        apiVersion="v1",
                        kind="Route",
                        metadata=dict(
                            name=route_name,
                            labels=labels.copy()
                        ),
                        spec=dict(
                            to=dict(
                                kind='Service',
                                name=name,
                            ),
                            port=dict(
                                targetPort="port-%s" % port
                            )
                        )
                    )
                    if hostname:
                        template['spec']['host'] = hostname
                else:
                    template = dict(
                        oso_route=OrderedDict(
                            project_name=self.project_name,
                            route_name=route_name,
                            labels=labels.copy(),
                            service_name=name,
                            service_port="port-%s" % port,
                            replace=True,
                        )
                    )
                    if hostname:
                        template['oso_route']['host'] = hostname

                    if state != 'present':
                        template['oso_route'] = state

                templates.append(template)

        return templates

    def _get_service_ports(self, service):
        '''
        Get the external port and the target container port.

        :param service: dict of service attributes
        :return: (external port, target port)
        '''

        #TODO: port ranges?

        result = []
        for port in service.get('ports', []):
            if isinstance(port, basestring) and ':' in port:
                parts = port.split(':')
                result.append(parts[0])
            else:
                result.append(port)
        return result



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
import os
import subprocess
import shlex
import select
import yaml

logger = logging.getLogger(__name__)


from ..exceptions import AnsibleContainerShipItException


def create_path(path):
    try:
        os.makedirs(path)
    except OSError:
        pass
    except Exception as exc:
        raise AnsibleContainerShipItException("Error creating %s - %s" % (path, str(exc)))


def run_command(args):
    '''
    Execute a command, returns rc, stdout, and stderr.

    :param args: Command to execute.
    '''

    if isinstance(args, basestring):
        if isinstance(args, unicode):
            args = args.encode('utf-8')
        args = shlex.split(args)
    else:
        raise AnsibleContainerShipItException("Argument 'args' to run_command must be list or string")

    args = [os.path.expandvars(os.path.expanduser(x)) for x in args if x is not None]

    st_in = None

    # Clean out python paths set by ziploader
    if 'PYTHONPATH' in os.environ:
        pypaths = os.environ['PYTHONPATH'].split(':')
        pypaths = [x for x in pypaths \
                    if not x.endswith('/ansible_modlib.zip') \
                    and not x.endswith('/debug_dir')]
        os.environ['PYTHONPATH'] = ':'.join(pypaths)

    kwargs = dict(
        executable=None,
        shell=False,
        close_fds=True,
        stdin=st_in,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    # store the pwd
    prev_dir = os.getcwd()

    try:
        cmd = subprocess.Popen(args, **kwargs)
        stdout = ''
        stderr = ''
        rpipes = [cmd.stdout, cmd.stderr]
        while True:
            rfd, wfd, efd = select.select(rpipes, [], rpipes, 1)
            if cmd.stdout in rfd:
                dat = os.read(cmd.stdout.fileno(), 9000)
                stdout += dat
                if dat == '':
                    rpipes.remove(cmd.stdout)
            if cmd.stderr in rfd:
                dat = os.read(cmd.stderr.fileno(), 9000)
                stderr += dat
                if dat == '':
                    rpipes.remove(cmd.stderr)
            # only break out if no pipes are left to read or
            # the pipes are completely read and
            # the process is terminated
            if (not rpipes or not rfd) and cmd.poll() is not None:
                break
            # No pipes are left to read but process is not yet terminated
            # Only then it is safe to wait for the process to be finished
            # NOTE: Actually cmd.poll() is always None here if rpipes is empty
            elif not rpipes and cmd.poll() is not None:
                cmd.wait()
                # The process is terminated. Since no pipes to read from are
                # left, there is no need to call select() again.
                break
        cmd.stdout.close()
        cmd.stderr.close()
        rc = cmd.returncode
    except Exception:
        raise

    os.chdir(prev_dir)
    return rc, stdout, stderr


def represent_odict(dump, tag, mapping, flow_style=None):
    '''
    https://gist.github.com/miracle2k/3184458
    Like BaseRepresenter.represent_mapping, but does not issue the sort().
    '''
    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dump.alias_key is not None:
        dump.represented_objects[dump.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = mapping.items()
    for item_key, item_value in mapping:
        node_key = dump.represent_data(item_key)
        node_value = dump.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and not node_key.style):
            best_style = False
        if not (isinstance(node_value, yaml.ScalarNode) and not node_value.style):
            best_style = False
        value.append((node_key, node_value))
    if flow_style is None:
        if dump.default_flow_style is not None:
            node.flow_style = dump.default_flow_style
        else:
            node.flow_style = best_style
    return node



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import sys
import copy

from StringIO import StringIO
from functools import wraps
from distutils import spawn

from compose.cli import main
from compose.cli.log_printer import LogPrinter, build_log_presenters

__all__ = ['teed_stdout',
           'which_docker',
           'config_to_compose']

# Don't try this at home, kids.

# *sigh* Okay, fine. Well...
# We need to tee the stdout to a StringIO buffer as well as to stdout.
# If you've got a better idea, I'm all ears.

class Tee(StringIO):
    def write(self, x):
        StringIO.write(self, x)
        sys.stdout.write(x)

    def flush(self):
        StringIO.flush(self)
        sys.stdout.flush()


def monkeypatch__log_printer_from_project(buffer):
    @wraps(main.log_printer_from_project)
    def __wrapped__(
            project,
            containers,
            monochrome,
            log_args,
            cascade_stop=False,
            event_stream=None,
    ):
        return LogPrinter(
            containers,
            build_log_presenters(project.service_names, monochrome),
            event_stream or project.events(),
            cascade_stop=cascade_stop,
            output=buffer,
            log_args=log_args)

    return __wrapped__


original__log_printer_from_project = main.log_printer_from_project


class TeedStdout(object):
    stdout = None

    def __enter__(self):
        self.stdout = StringIO()
        main.log_printer_from_project = monkeypatch__log_printer_from_project(self.stdout)
        return self.stdout

    def __exit__(self, exc_type, exc_val, exc_tb):
        main.log_printer_from_project = original__log_printer_from_project

teed_stdout = TeedStdout

# Phew. That sucked.

def which_docker():
    return spawn.find_executable('docker')

def config_to_compose(config):
    # This could probably be better done - include what keys are in compose vs
    # removing the ones that aren't.
    compose = copy.deepcopy(config.get('services') or {})
    assert compose is not config.get('services')
    for service, service_config in compose.items():
        if 'options' in service_config:
            del service_config['options']
    logger.debug('Compose derived from config:')
    logger.debug(unicode(compose))
    return compose



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from collections import OrderedDict


import logging

logger = logging.getLogger(__name__)


class Service(object):

    def __init__(self, config=None, project_name=None):
        self.project_name = project_name
        self.config = config

    def get_template(self):
        return self._get_task_or_config(request_type="config")

    def get_task(self):
        return self._get_task_or_config(request_type="task")

    def _get_task_or_config(self, request_type="task"):
        templates = []
        for name, service in self.config.get('services', {}).items():
            new_service = self._create(request_type, name, service)
            if new_service:
                templates.append(new_service)
        return templates

    def _create(self, type, name, service):
        '''
        Generate an Openshift service configuration or playbook task.
        '''
        template = {}
        ports = self._get_ports(service)
        options = service.get('options', {}).get('openshift', {})
        state = options.get('state', 'present')

        if ports:
            labels = dict(
                app=self.project_name,
                service=name
            )
            if type == 'config' and state != 'absent':
                template = dict(
                    apiVersion="v1",
                    kind="Service",
                    metadata=dict(
                        name=name,
                        labels=labels.copy()
                    ),
                    spec=dict(
                        selector=labels.copy(),
                        ports=ports,
                    )
                )
            elif type == 'task':
                template = dict(
                    oso_service=OrderedDict(
                        project_name=self.project_name,
                        service_name=name,
                        labels=labels.copy(),
                        ports=ports,
                        selector=labels.copy()
                    )
                )
                if state != 'present':
                    template['oso_service']['state'] = state

        return template

    def _get_ports(self, service):
        # TODO - handle port ranges
        ports = []
        for port in service.get('ports', []):
            if isinstance(port, str) and ':' in port:
                parts = port.split(':')
                if not self._port_in_list(parts[0], ports):
                    ports.append(dict(port=int(parts[0]), targetPort=int(parts[1]), name='port-%s' % parts[0]))
            else:
                if not self._port_in_list(port, ports):
                    ports.append(dict(port=int(port), targetPort=int(port), name='port-%s' % port))

        for port in service.get('expose', []):
            if not self._port_in_list(port, ports):
                ports.append(dict(port=int(port), targetPort=int(port), name='port-%s' % port))
        return ports

    @staticmethod
    def _port_in_list(port, ports):
        found = False
        for p in ports:
            if p['port'] == int(port):
                found = True
                break
        return found




# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import os
import yaml
import re
import glob

from datetime import datetime

from .constants import SHIPIT_PATH, SHIPIT_PLAYBOOK_PREFIX, SHIPIT_ROLES_DIR
from collections import OrderedDict
from .utils import create_path, represent_odict
from ..utils import jinja_render_to_temp


class BaseShipItEngine(object):
    name = None

    def __init__(self, base_path=None, project_name=None, config=None):
        self.base_path = base_path
        self.project_name = project_name
        self.config = config
        self.role_name = "%s-%s" % (self.project_name, self.name)
        self.roles_path = os.path.join(self.base_path, SHIPIT_PATH, SHIPIT_ROLES_DIR, self.role_name)

    def add_options(self, subparser):
        """
        Given an argument subparser, add to it the arguments and options
        this engine allows.

        https://docs.python.org/2/library/argparse.html#sub-commands

        :param subparser: An argparse.ArgumentSubparser
        :return: None
        """
        subparser.add_argument('--save-config', action='store_true',
                               help=(u'Generate and save the %s configuration files in '
                                     u'ansible/shipit_config/kubernetes.' % self.name),
                               dest='save_config', default=False)

        subparser.add_argument('--pull-from', action='store',
                               help=u'Name of a registry defined in container.yml or the actual URL the cluster will '
                                    u'use to pull images. If passing a URL, an example would be: '
                                    u'"registry.example.com:5000/myproject"',
                               dest='pull_from', default=None)

    def run(self):
        """
        Generate an Ansible role and sample playbook to deploy to the target cloud provider.

        :return:
        """
        raise NotImplementedError()

    def save_config(self):
        """
        Called when --save-config option True. Generate configuration templates and write to filesystem.

        :return: string: name of directory where config was saved
        """
        raise NotImplementedError()

    def _copy_modules(self):
        '''
        Copy cloud ansible modules to role library path.
        '''
        cls_dir = os.path.dirname(os.path.realpath(__file__))
        modules_dir = os.path.join(cls_dir, self.name, 'modules')
        library_path = os.path.join(self.roles_path, 'library')
        logger.debug("Copying modules from %s to %s" % (modules_dir, library_path))
        create_path(library_path)

        include_files = []
        for mod in glob.glob(modules_dir + '/*.py'):
            with open(mod, 'r') as mod_file:
                for line in mod_file:
                    match = re.search('#include--> ?(\w+.py)', line)
                    if match and match.groups():
                        include_files += list(match.groups())

        include_files = list(set(include_files))
        for mod in glob.glob(modules_dir + '/*.py'):
            base_file = os.path.basename(mod)
            if base_file not in include_files and not base_file.endswith('__init__.py'):
                with open(os.path.join(library_path, base_file), 'w') as new_file:
                    with open(mod, 'r') as mod_file:
                        for line in mod_file:
                            match = re.search('#include--> ?(\w+.py)', line)
                            if match and match.groups():
                                with open(os.path.join(modules_dir, match.group(1)), 'r') as inc:
                                    inc_contents = inc.read()
                                new_file.write('\n')
                                new_file.write(inc_contents)
                            else:
                                new_file.write(line)

    def init_role(self):
        '''
        If it does not exist, create the role directory and template in the initial files.
        If it does, backup anything that we plan to overwrite.

        :return:
        '''
        # Init the role directory without overwriting any existing files.
        logger.debug("Creating role path %s" % self.roles_path)
        create_path(self.roles_path)

        shipit_role_paths = {
            u'base': [u'README.j2', u'travis.j2.yml'],
            u'defaults': [u'defaults.j2.yml'],
            u'meta': [u'meta.j2.yml'],
            u'test': [u'test.j2.yml', u'travis.j2.yml'],
            u'tasks': [],
        }

        context = {
            u'role_name': self.role_name,
            u'project_name': self.project_name,
            u'shipit_engine_name': self.name
        }

        for path, templates in shipit_role_paths.items():
            role_dir = os.path.join(self.roles_path, path if path != 'base' else '')
            if path != 'base':
                create_path(role_dir)
            for template in templates:
                target_name = template.replace('.j2', '')
                if target_name.startswith('travis'):
                    target_name = '.' + target_name
                if not os.path.exists(os.path.join(role_dir, target_name)):
                    logger.debug("Rendering template for %s/%s" % (path, template))
                    jinja_render_to_temp('shipit_role/%s' % template,
                                         role_dir,
                                         target_name,
                                         **context)

        # if tasks/main.yml exists, back it up
        now = datetime.today().strftime('%y%m%d%H%M%S')
        tasks_file = os.path.join(self.roles_path, 'tasks', 'main.yml')
        new_tasks_file = os.path.join(self.roles_path, 'tasks', 'main_%s.yml' % now)

        if os.path.exists(tasks_file):
            logger.debug("Backing up tasks/main.yml to main_%s.yml" % now)
            os.rename(tasks_file, new_tasks_file)

        #TODO: limit the number of backups kept?


    def create_role(self, tasks):
        '''
        Creates tasks/main.yml with all the deployment tasks, and copy modules into the library dir.

        :param tasks: dict of playbook tasks to be rendered in yaml
        :return: None
        '''

        yaml.SafeDumper.add_representer(OrderedDict,
                                        lambda dumper, value: represent_odict(dumper, u'tag:yaml.org,2002:map', value))

        # Create tasks/main.yml
        output_tasks = []
        for task in tasks:
            task['register'] = "output"
            output_tasks.append(task)
            output_tasks.append(dict(
                debug="var=output",
                when="playbook_debug"
            ))
        stream = yaml.safe_dump(output_tasks, default_flow_style=False)
        with open(os.path.join(self.roles_path, 'tasks', 'main.yml'), 'w') as f:
            f.write(re.sub(r'^-', u'\n-', stream, flags=re.M))

        # Copy ansible modules to library path
        self._copy_modules()

    def create_playbook(self):
        '''
        Create a simple playbook to execute the role. Will only create the playbook if one
        does not exist.

        :return: None
        '''
        playbook_name = "%s-%s.yml" % (SHIPIT_PLAYBOOK_PREFIX, self.name)
        playbook_path = os.path.join(self.base_path, SHIPIT_PATH, playbook_name)
        if not os.path.exists(playbook_path):
            logger.debug('Creating the sample playbook')
            play = OrderedDict()
            play['name'] = "Deploy %s to  %s" % (self.project_name, self.name)
            play['hosts'] = 'localhost'
            play['gather_facts'] = False
            play['connection'] = 'local'
            play['vars'] = dict(
                playbook_debug=False
            )
            play['roles'] = [dict(
                role=self.role_name
            )]
            with open(playbook_path, 'w') as f:
                f.write(yaml.safe_dump([play], default_flow_style=False))




#!/usr/bin/python
#
# Copyright 2016 Ansible by Red Hat
#
# This file is part of ansible-container
#

DOCUMENTATION = '''

module: oso_service

short_description: Create or remove a service on a Kubernetes cluster.

description:
  - Create or remove a service on a Kubernetes cluster by setting the C(state) to I(present) or I(absent).
  - The module is idempotent and will not replace an existing service unless the C(reload) option is passed.
  - Supports check mode. Use check mode to view a list of actions the module will take.

options:

'''

EXAMPLES = '''
'''

RETURN = '''
'''
import logging
import logging.config

from ansible.module_utils.basic import *

logger = logging.getLogger('kube_service')

LOGGING = (
    {
        'version': 1,
        'disable_existing_loggers': True,
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
            },
            'file': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': 'ansible-container.log'
            }
        },
        'loggers': {
            'kube_service': {
                'handlers': ['file'],
                'level': 'INFO',
            },
            'container': {
                'handlers': ['file'],
                'level': 'INFO',
            },
            'compose': {
                'handlers': [],
                'level': 'INFO'
            },
            'docker': {
                'handlers': [],
                'level': 'INFO'
            }
        },
    }
)


class KubeServiceManager(object):

    def __init__(self):

        self.arg_spec = dict(
            state=dict(type='str', choices=['present', 'absent'], default='present'),
            labels=dict(type='dict'),
            ports=dict(type='list', required=True),
            service_name=dict(type='str', required=True),
            loadbalancer=dict(type='bool', default=False),
            replace=dict(type='bool', default=False),
            selector=dict(type='dict', required=True),
            type=dict(type='str', choices=['ClusterIP', 'LoadBalancer', 'NodePort'], default='NodePort'),
        )

        self.module = AnsibleModule(self.arg_spec,
                                    supports_check_mode=True)

        self.state = None
        self.labels = None
        self.ports = None
        self.service_name = None
        self.loadbalancer = None
        self.selector = None
        self.replace = None
        self.api = None
        self.type = None
        self.check_mode = self.module.check_mode
        self.debug = self.module._debug

    def exec_module(self):

        for key in self.arg_spec:
            setattr(self, key, self.module.params.get(key))

        if self.debug:
            LOGGING['loggers']['container']['level'] = 'DEBUG'
            LOGGING['loggers']['kube_service']['level'] = 'DEBUG'
        logging.config.dictConfig(LOGGING)

        self.api = KubeAPI(self.module)

        actions = []
        changed = False
        services = dict()
        results = dict()

        if self.state == 'present':
            service = self.api.get_resource('service', self.service_name)
            if not service:
                template = self._create_template()
                changed = True
                actions.append("Create service %s" % self.service_name)
                if not self.check_mode:
                    self.api.create_from_template(template=template)
            elif service and self.replace:
                template = self._create_template()
                changed = True
                actions.append("Replace service %s" % self.service_name)
                if not self.check_mode:
                    self.api.replace_from_template(template=template)

            services[self.service_name.replace('-', '_') + '_service'] = \
                self.api.get_resource('service', self.service_name)

        elif self.state == 'absent':
            if self.api.get_resource('service', self.service_name):
                changed = True
                actions.append("Delete service %s" % self.service_name)
                if not self.check_mode:
                    self.api.delete_resource('service', self.service_name)

        results['changed'] = changed

        if self.check_mode:
            results['actions'] = actions

        if services:
            results['ansible_facts'] = services

        self.module.exit_json(**results)

    def _create_template(self):
        '''
        apiVersion: v1
            kind: Service
            metadata:
              name: frontend
              labels:
                app: guestbook
                tier: frontend
            spec:
              # if your cluster supports it, uncomment the following to automatically create
              # an external load-balanced IP for the frontend service.
              # type: LoadBalancer
              ports:
                # the port that this service should serve on
              - port: 80
              selector:
                app: guestbook
                tier: frontend
        '''

        selector = self.selector if self.selector else self.service_name
        self._update_ports(self.ports)

        template = dict(
            apiVersion="v1",
            kind="Service",
            metadata=dict(
                name=self.service_name
            ),
            spec=dict(
                selector=selector,
                ports=self.ports,
                type=self.type
            )
        )

        if self.labels:
            template['metadata']['labels'] = self.labels

        return template

    @staticmethod
    def _update_ports(ports):
        for port in ports:
            if not port.get('name'):
                port['name'] = "port_%s" % port['port']
            if not port.get('protocol'):
                port['protocol'] = "TCP"

#The following will be included by `ansble-container shipit` when cloud modules are copied into the role library path.
#include--> kube_api.py


def main():
    manager = KubeServiceManager()
    manager.exec_module()

if __name__ == '__main__':
    main()



import re
import json


class OriginAPI(object):

    def __init__(self, module, target="oc"):
        self.target = target
        self.module = module

    @staticmethod
    def use_multiple_deployments(services):
        '''
        Inspect services and return True if the app supports multiple replica sets.

        :param services: list of docker-compose service dicts
        :return: bool
        '''
        multiple = True
        for service in services:
            if not service.get('ports'):
                multiple = False
            if service.get('volumes_from'):
                multiple = False
        return multiple

    def call_api(self, cmd, data=None, check_rc=False, error_msg=None):
        rc, stdout, stderr = self.module.run_command(cmd, data=data)
        logger.debug("Received rc: %s" % rc)
        logger.debug("stdout:")
        logger.debug(stdout)
        logger.debug("stderr:")
        logger.debug(stderr)

        if check_rc and rc != 0:
            self.module.fail_json(msg=error_msg, stderr=stderr, stdout=stdout)

        return rc, stdout, stderr

    def create_from_template(self, template=None, template_path=None):
        if template_path:
            logger.debug("Create from template %s" % template_path)
            error_msg = "Error Creating %s" % template_path
            cmd = "%s create -f %s" % (self.target, template_path)
            rc, stdout, stderr = self.call_api(cmd, check_rc=True, error_msg=error_msg)
            return stdout

        if template:
            logger.debug("Create from template:")
            formatted_template = json.dumps(template, sort_keys=False, indent=4, separators=(',', ':'))
            logger.debug(formatted_template)
            cmd = "%s create -f -" % self.target
            rc, stdout, stderr = self.call_api(cmd, data=formatted_template, check_rc=True,
                                               error_msg="Error creating from template.")
            return stdout

    def replace_from_template(self, template=None, template_path=None):
        if template_path:
            logger.debug("Replace from template %s" % template_path)
            cmd = "%s replace -f %s" % (self.target, template_path)
            error_msg = "Error replacing %s" % template_path
            rc, stdout, stderr = self.call_api(cmd, check_rc=True, error_msg=error_msg)
            return stdout
        if template:
            logger.debug("Replace from template:")
            formatted_template = json.dumps(template, sort_keys=False, indent=4, separators=(',', ':'))
            logger.debug(formatted_template)
            cmd = "%s replace -f -" % self.target
            rc, stdout, stderr = self.call_api(cmd, data=formatted_template, check_rc=True,
                                               error_msg="Error replacing from template")
            return stdout

    def delete_resource(self, type, name):
        cmd = "%s delete %s/%s" % (self.target, type, name)
        logger.debug("exec: %s" % cmd)
        error_msg = "Error deleting %s/%s" % (type, name)
        rc, stdout, stderr = self.call_api(cmd, check_rc=True, error_msg=error_msg)
        return stdout

    def get_resource(self, type, name):
        result = None
        cmd = "%s get %s/%s -o json" % (self.target, type, name)
        logger.debug("exec: %s" % cmd)
        rc, stdout, stderr = self.call_api(cmd)
        if rc == 0:
            result = json.loads(stdout) 
        elif rc != 0 and not re.search('not found', stderr):
            error_msg = "Error getting %s/%s" % (type, name)
            self.module.fail_json(msg=error_msg, stderr=stderr, stdout=stdout)
        return result
   
    def set_context(self, context_name):
        cmd = "%s user-context %s" % (self.target, context_name)
        logger.debug("exec: %s" % cmd)
        error_msg = "Error switching to context %s" % context_name
        rc, stdout, stderr = self.call_api(cmd, check_rc=True, error_msg=error_msg)
        return stdout

    def set_project(self, project_name):
        result = True
        cmd = "%s project %s" % (self.target, project_name)
        logger.debug("exec: %s" % cmd)
        rc, stdout, stderr = self.call_api(cmd)
        if rc != 0:
            result = False
            if not re.search('does not exist', stderr):
                error_msg = "Error switching to project %s" % project_name
                self.module.fail_json(msg=error_msg, stderr=stderr, stdout=stdout)
        return result

    def create_project(self, project_name):
        result = True
        cmd = "%s new-project %s" % (self.target, project_name)
        logger.debug("exec: %s" % cmd)
        error_msg = "Error creating project %s" % project_name
        self.call_api(cmd, check_rc=True, error_msg=error_msg)
        return result

    def get_deployment(self, deployment_name):
        cmd = "%s deploy %s" % (self.target, deployment_name)
        logger.debug("exec: %s" % cmd)
        rc, stdout, stderr = self.call_api(cmd)
        if rc != 0:
            if not re.search('not found', stderr):
                error_msg = "Error getting deployment state %s" % deployment_name
                self.module.fail_json(msg=error_msg, stderr=stderr, stdout=stdout)
        return stdout



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import os
import re
import tarfile
import getpass
import json
import base64
import pprint

import docker
from docker.client import errors as docker_errors
from docker.utils import kwargs_from_env
from compose.cli.command import project_from_options
from compose.cli import main
from yaml import dump as yaml_dump

from ..exceptions import (AnsibleContainerNotInitializedException,
                          AnsibleContainerNoAuthenticationProvidedException,
                          AnsibleContainerDockerConfigFileException,
                          AnsibleContainerDockerLoginException,
                          AnsibleContainerNoMatchingHosts)

from ..engine import BaseEngine, REMOVE_HTTP
from ..utils import *
from .. import __version__ as release_version
from .utils import *

if not os.environ.get('DOCKER_HOST'):
    logger.warning('No DOCKER_HOST environment variable found. Assuming UNIX '
                   'socket at /var/run/docker.sock')


class Engine(BaseEngine):

    engine_name = 'Docker'
    orchestrator_name = 'Docker Compose'
    builder_container_img_name = 'ansible-container'
    builder_container_img_tag = 'ansible-container-builder'
    default_registry_url = 'https://index.docker.io/v1/'
    default_registry_name = 'dockerhub'
    _client = None
    _orchestrated_hosts = None
    api_version = ''
    temp_dir = None

    def all_hosts_in_orchestration(self):
        """
        List all hosts being orchestrated by the compose engine.

        :return: list of strings
        """
        return (self.config.get('services') or {}).keys()

    def hosts_touched_by_playbook(self):
        """
        List all hosts touched by the execution of the build playbook.

        :return: list of strings
        """
        if not self._orchestrated_hosts:
            with teed_stdout() as stdout, make_temp_dir() as temp_dir:
                self.orchestrate('listhosts', temp_dir,
                                 hosts=[self.builder_container_img_name])
                logger.info('Cleaning up Ansible Container builder...')
                builder_container_id = self.get_builder_container_id()
                self.remove_container_by_id(builder_container_id)
                # We need to cleverly extract the host names from the output...
                logger.debug('--list-hosts\n%s', stdout.getvalue())
                lines = stdout.getvalue().split('\r\n')
                lines_minus_builder_host = [line.rsplit('|', 1)[1] for line
                                            in lines if '|' in line]
                host_lines = [line for line in lines_minus_builder_host
                              if line.startswith('       ')]
                self._orchestrated_hosts = list(set([line.strip() for line in host_lines]))
        return filter(None, self._orchestrated_hosts)

    def build_buildcontainer_image(self):
        """
        Build in the container engine the builder container

        :return: generator of strings
        """
        assert_initialized(self.base_path)
        client = self.get_client()
        with make_temp_dir() as temp_dir:
            logger.info('Building Docker Engine context...')
            tarball_path = os.path.join(temp_dir, 'context.tar')
            tarball_file = open(tarball_path, 'wb')
            tarball = tarfile.TarFile(fileobj=tarball_file,
                                      mode='w')
            container_dir = os.path.normpath(os.path.join(self.base_path,
                                                          'ansible'))
            try:
                tarball.add(container_dir, arcname='ansible')
            except OSError:
                raise AnsibleContainerNotInitializedException()
            jinja_render_to_temp('ansible-dockerfile.j2', temp_dir,
                                 'Dockerfile')
            tarball.add(os.path.join(temp_dir, 'Dockerfile'),
                        arcname='Dockerfile')

            for context_file in ['builder.sh', 'ansible-container-inventory.py',
                                 'ansible.cfg', 'wait_on_host.py', 'ac_galaxy.py']:
                tarball.add(os.path.join(jinja_template_path(), context_file),
                            arcname=context_file)

            tarball.close()
            tarball_file.close()
            tarball_file = open(tarball_path, 'rb')
            logger.info('Starting Docker build of Ansible Container image (please be patient)...')
            return client.build(fileobj=tarball_file,
                                custom_context=True,
                                tag=self.builder_container_img_tag,
                                nocache=True,
                                rm=True)

    def get_image_id_by_tag(self, name):
        """
        Query the engine to get an image identifier by tag

        :param name: the image name
        :return: the image identifier
        """
        client = self.get_client()
        try:
            return client.images(name=name, quiet=True)[0]
        except IndexError:
            raise NameError('No image with the name %s' % name)

    def get_container_id_by_name(self, name):
        """
        Query the engine to get a container identifier by name

        :param name: the container name
        :return: the container identifier
        """
        client = self.get_client()
        try:
            return client.containers(
                filters={'name': name},
                limit=1, all=True, quiet=True)[0]
        except IndexError:
            raise NameError('No container with the name %s' % name)

    def remove_container_by_name(self, name):
        """
        Remove a container from the engine given its name

        :param name: the name of the container to remove
        :return: None
        """
        client = self.get_client()
        container_id, = client.containers(
            filters={'name': 'ansible_%s_1' % name},
            limit=1, all=True, quiet=True
        )
        self.remove_container_by_id(container_id)

    def remove_container_by_id(self, id):
        """
        Remove a container from the engine given its identifier

        :param id: container identifier
        :return: None
        """
        client = self.get_client()
        client.remove_container(id)

    def get_builder_image_id(self):
        """
        Query the enginer to get the builder image identifier

        :return: the image identifier
        """
        return self.get_image_id_by_tag(self.builder_container_img_tag)

    def get_builder_container_id(self):
        """
        Query the enginer to get the builder container identifier

        :return: the container identifier
        """
        return self.get_container_id_by_name(self.builder_container_img_name)

    def build_was_successful(self):
        """
        After the build was complete, did the build run successfully?

        :return: bool
        """
        client = self.get_client()
        build_container_info, = client.containers(
            filters={'name': 'ansible_ansible-container_1'},
            limit=1, all=True
        )
        # Not the best way to test for success or failure, but it works.
        exit_status = build_container_info['Status']
        return '(0)' in exit_status

    def get_config_for_shipit(self, pull_from=None, url=None, namespace=None):
        '''
        Retrieve the configuration needed to run the shipit command

        :param pull_from: the exact registry URL to use
        :param url: registry url. required, if pull_from not provided.
        :param namespace: path to append to the url. required if pull_from not provided.
        :return: config dict
        '''
        config = get_config(self.base_path, var_file=self.var_file)
        client = self.get_client()

        if pull_from:
            image_path = re.sub(r'/$', '', REMOVE_HTTP.sub('', pull_from))
        else:
            image_path = namespace
            if url != self.default_registry_url:
                url = REMOVE_HTTP.sub('', url)
                image_path = "%s/%s" % (re.sub(r'/$', '', url), image_path)

        logger.info("Images will be pulled from %s" % image_path)
        orchestrated_hosts = self.hosts_touched_by_playbook()
        for host, service_config in config.get('services', {}).items():
            if host in orchestrated_hosts:
                image_id, image_buildstamp = get_latest_image_for(self.project_name, host, client)
                service_config.update(
                    dict(
                        image='%s/%s-%s:%s' % (image_path, self.project_name, host, image_buildstamp)
                    )
                )
        return config

    # Docker-compose uses docopt, which outputs things like the below
    # So I'm starting with the defaults and then updating them.
    # One structure for the global options and one for the command specific

    DEFAULT_COMPOSE_OPTIONS = {
        u'--help': False,
        u'--host': None,
        u'--project-name': None,
        u'--skip-hostname-check': False,
        u'--tls': False,
        u'--tlscacert': None,
        u'--tlscert': None,
        u'--tlskey': None,
        u'--tlsverify': False,
        u'--verbose': False,
        u'--version': False,
        u'-h': False,
        u'--file': [],
        u'COMMAND': None,
        u'ARGS': []
    }

    DEFAULT_COMPOSE_UP_OPTIONS = {
        u'--abort-on-container-exit': False,
        u'--build': False,
        u'--force-recreate': False,
        u'--no-color': False,
        u'--no-deps': False,
        u'--no-recreate': False,
        u'--no-build': False,
        u'--remove-orphans': False,
        u'--timeout': None,
        u'-d': False,
        u'SERVICE': []
    }

    DEFAULT_COMPOSE_STOP_OPTIONS = {
        u'--timeout': None,
        u'SERVICE': []
    }

    DEFAULT_COMPOSE_RESTART_OPTIONS = {
        u'--timeout': None,
        u'SERVICE': []
    }

    def orchestrate(self, operation, temp_dir, hosts=[], context={}):
        """
        Execute the compose engine.

        :param operation: One of build, run, or listhosts
        :param temp_dir: A temporary directory usable as workspace
        :param hosts: (optional) A list of hosts to limit orchestration to
        :return: The exit status of the builder container (None if it wasn't run)
        """
        is_detached = self.params.pop('detached', False)
        try:
            builder_img_id = self.get_image_id_by_tag(
                self.builder_container_img_tag)
        except NameError:
            image_version = '.'.join(release_version.split('.')[:2])
            builder_img_id = 'ansible/%s:%s' % (
                self.builder_container_img_tag, image_version)

        options, command_options, command = self.bootstrap_env(
            temp_dir=temp_dir,
            builder_img_id=builder_img_id,
            behavior='orchestrate',
            compose_option='up',
            operation=operation,
            context=context)

        options.update({
            u'COMMAND': 'up',
            u'ARGS': ['--no-build'] + hosts})

        command_options[u'--no-build'] = True
        command_options[u'SERVICE'] = hosts
        command_options[u'--remove-orphans'] = self.params.get(
            'remove_orphans', False)

        if is_detached:
            logger.info('Deploying application in detached mode')
            command_options[u'-d'] = True

        command.up(command_options)

    def orchestrate_build_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during build.

        :return: dictionary
        """
        return {'--abort-on-container-exit': True,
                '--force-recreate': True}

    def orchestrate_run_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during run.

        :return: dictionary
        """
        return {}

    def terminate_stop_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during stop.

        :return: dictionary
        """
        return {}

    def orchestrate_install_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during install calls.

        :return: dictionary
        """
        return {}

    def orchestrate_listhosts_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during listhosts.

        :return: dictionary
        """
        return {'--no-color': True}

    def terminate(self, operation, temp_dir, hosts=[]):

        options, command_options, command = self.bootstrap_env(
            temp_dir=temp_dir,
            behavior='terminate',
            operation=operation,
            compose_option='stop'
        )

        options.update({u'COMMAND': 'stop'})

        command_options[u'SERVICE'] = hosts

        if self.params.get('force'):
            command.kill(command_options)
        else:
            command.stop(command_options)

    def restart(self, operation, temp_dir, hosts=[]):

        options, command_options, command = self.bootstrap_env(
            temp_dir=temp_dir,
            behavior='restart',
            operation=operation,
            compose_option='restart'
        )

        options.update({u'COMMAND': 'restart'})

        command_options[u'SERVICE'] = hosts

        command.restart(command_options)

    def restart_restart_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during restart.

        :return: dictionary
        """
        return {}

    def get_config_for_restart(self):
        compose_config = config_to_compose(self.config)
        return compose_config

    def _fix_volumes(self, service_name, service_config):
        # If there are volumes defined for this host, we need to create the
        # volume if one doesn't already exist.
        client = self.get_client()
        for volume in service_config.get('volumes', []):
            if ':' not in volume:
                # This is an unnamed volume. We have to handle making this
                # volume with a predictable name.
                volume_name = '%s-%s-%s' % (self.project_name,
                                            service_name,
                                            volume.replace('/', '_'))
                try:
                    client.inspect_volume(name=volume_name)
                except docker_errors.NotFound, e:
                    # We need to create this volume
                    client.create_volume(name=volume_name, driver='local')
                service_config['volumes'].remove(volume)
                service_config['volumes'].append('%s:%s' % (volume_name, volume))

    def get_config_for_build(self):
        compose_config = config_to_compose(self.config)
        orchestrated_hosts = self.hosts_touched_by_playbook()
        if self.params.get('service'):
            # only build a subset of the orchestrated hosts
            orchestrated_hosts = list(set(orchestrated_hosts).intersection(self.params['service']))
            for host in set(compose_config.keys()) - set(orchestrated_hosts):
                del compose_config[host]
            if not compose_config:
                raise AnsibleContainerNoMatchingHosts()
        logger.debug('Orchestrated hosts: %s', orchestrated_hosts)

        for service, service_config in compose_config.items():
            if service in orchestrated_hosts:
                logger.debug('Setting %s to sleep', service)
                service_config.update(
                    dict(
                        user='root',
                        working_dir='/',
                        command='sh -c "while true; do sleep 1; done"',
                        entrypoint=[],
                    )
                )
                # Set ANSIBLE_CONTAINER=1 in env
                if service_config.get('environment'):
                    if isinstance(service_config['environment'], list):
                        service_config['environment'].append("ANSIBLE_CONTAINER=1")
                    elif isinstance(service_config['environment'], dict):
                        service_config['environment']['ANSIBLE_CONTAINER'] = 1
                else:
                    service_config['environment'] = dict(ANSIBLE_CONTAINER=1)

            if not self.params['rebuild']:
                tag = '%s-%s:latest' % (self.project_name,
                                        service)
                try:
                    self.get_image_id_by_tag(tag)
                except NameError:
                    logger.info('No image found for tag %s, so building from scratch',
                                '%s-%s:latest' % (self.project_name, service))
                    # have to rebuild this from scratch, as the image doesn't
                    # exist in the engine
                    pass
                else:
                    logger.debug('No NameError raised when searching for tag %s',
                                 '%s-%s:latest' % (self.project_name, service))
                    service_config['image'] = tag
            self._fix_volumes(service, service_config)
        return compose_config

    def get_config_for_run(self):
        if not self.params['production']:
            self.config.set_env('dev')
        compose_config = config_to_compose(self.config)
        orchestrated_hosts = self.hosts_touched_by_playbook()
        for service, service_config in compose_config.items():
            if service in orchestrated_hosts:
                service_config.update(
                    dict(
                        image='%s-%s:latest' % (self.project_name, service)
                    )
                )
            self._fix_volumes(service, service_config)
        return compose_config

    def get_config_for_stop(self):
        compose_config = config_to_compose(self.config)
        return compose_config

    def get_config_for_listhosts(self):
        compose_config = config_to_compose(self.config)
        for service, service_config in compose_config.items():
            service_config.update(
                dict(
                    user='root',
                    working_dir='/',
                    command='sh -c "while true; do sleep 1; done"',
                    entrypoint=[]
                )
            )
        return compose_config

    def get_config_for_install(self):
        compose_config = config_to_compose(self.config)
        return compose_config

    def post_build(self, host, version, flatten=True, purge_last=True):
        client = self.get_client()
        container_id, = client.containers(
            filters={'name': 'ansible_%s_1' % host},
            limit=1, all=True, quiet=True
        )
        previous_image_id, previous_image_buildstamp = get_latest_image_for(
            self.project_name, host, client
        )
        cmd = self.config['services'][host].get('command', '')
        if isinstance(cmd, list):
            cmd = json.dumps(cmd)
        entrypoint = self.config['services'][host].get('entrypoint', '')
        if isinstance(entrypoint, list):
            entrypoint = json.dumps(entrypoint)
        image_config = dict(
            USER=self.config['services'][host].get('user', 'root'),
            WORKDIR=self.config['services'][host].get('working_dir', '/'),
            LABEL='com.docker.compose.oneoff="" com.docker.compose.project="%s"' % self.project_name,
            ENTRYPOINT=entrypoint,
            CMD=cmd
        )
        if flatten:
            logger.info('Flattening image...')
            exported = client.export(container_id)
            client.import_image_from_data(
                exported.read(),
                repository='%s-%s' % (self.project_name, host),
                tag=version)
        else:
            logger.info('Committing image...')
            client.commit(container_id,
                          repository='%s-%s' % (self.project_name, host),
                          tag=version,
                          message='Built using Ansible Container',
                          changes=u'\n'.join(
                              [u'%s %s' % (k, unicode(v))
                               for k, v in image_config.items()]
                          ))
        image_id, = client.images(
            '%s-%s:%s' % (self.project_name, host, version),
            quiet=True
        )
        logger.info('Exported %s-%s with image ID %s', self.project_name, host,
                    image_id)
        client.tag(image_id, '%s-%s' % (self.project_name, host), tag='latest',
                   force=True)
        logger.info('Cleaning up %s build container...', host)
        client.remove_container(container_id)

        image_data = client.inspect_image(image_id)
        parent_sha = ''
        if image_data:
            parent_sha = image_data.get('Parent', '')

        if purge_last and previous_image_id and previous_image_id not in parent_sha:
            logger.info('Removing previous image...')
            client.remove_image(previous_image_id, force=True)

    DEFAULT_CONFIG_PATH = '~/.docker/config.json'

    def registry_login(self, username=None, password=None, email=None, url=None, config_path=DEFAULT_CONFIG_PATH):
        """
        Logs into registry for this engine

        :param username: Username to login with
        :param password: Password to login with - None to prompt user
        :param email: Email address to login with
        :param url: URL of registry - default defined per backend
        :param config_path: path to the registry config file
        :return: None
        """

        #TODO Make config_path a command line option?

        client = self.get_client()
        if not url:
            url = self.default_registry_url

        if username:
            # We assume if no username was given, the docker config file
            # suffices
            while not password:
                password = getpass.getpass(u'Enter password for %s at %s: ' % (
                    username, url
                ))
            try:
                client.login(username=username, password=password, email=email,
                             registry=url)
            except Exception as exc:
                raise AnsibleContainerDockerLoginException("Error logging into registry: %s" % str(exc))

            self.update_config_file(username, password, email, url, config_path)

        username, email = self.currently_logged_in_registry_user(url)
        if not username:
            raise AnsibleContainerNoAuthenticationProvidedException(
                u'Please provide login credentials for registry %s.' % url)
        return username

    DOCKER_CONFIG_FILEPATH_CASCADE = [
        os.environ.get('DOCKER_CONFIG', ''),
        os.path.join(os.environ.get('HOME', ''), '.docker', 'config.json'),
        os.path.join(os.environ.get('HOME', ''), '.dockercfg')
    ]

    def currently_logged_in_registry_user(self, url):
        """
        Gets logged in user from configuration for a URL for the registry for
        this engine.

        :param url: URL
        :return: (username, email) tuple
        """
        username = None
        docker_config = None
        for docker_config_filepath in self.DOCKER_CONFIG_FILEPATH_CASCADE:
            if docker_config_filepath and os.path.exists(docker_config_filepath):
                docker_config = json.load(open(docker_config_filepath))
                break
        if not docker_config:
            raise AnsibleContainerDockerConfigFileException("Unable to read your docker config file. Try providing "
                                                            "login credentials for the registry.")
        if 'auths' in docker_config:
            docker_config = docker_config['auths']
        auth_key = docker_config.get(url, {}).get('auth', '')
        email = docker_config.get(url, {}).get('email', '')
        if auth_key:
            username, password = base64.decodestring(auth_key).split(':', 1)
        return username, email

    def update_config_file(self, username, password, email, url, config_path):
        '''
        Update the config file with the authorization.

        :param username:
        :param password:
        :param email:
        :param url:
        :param config_path
        :return: None
        '''

        path = os.path.expanduser(config_path)
        if not os.path.exists(path):
            # Create an empty config, if none exists
            config_path_dir = os.path.dirname(path)
            if not os.path.exists(config_path_dir):
                try:
                    os.makedirs(config_path_dir)
                except Exception as exc:
                    raise AnsibleContainerDockerConfigFileException("Failed to create file %s - %s" %
                                                                    (config_path_dir, str(exc)))
            self.write_config(path, dict(auths=dict()))

        try:
            # read the existing config
            config = json.load(open(path, "r"))
        except ValueError:
            config = dict()

        if not config.get('auths'):
            config['auths'] = dict()

        if not config['auths'].get(url):
            config['auths'][url] = dict()

        encoded_credentials = dict(
            auth=base64.b64encode(username + b':' + password),
            email=email
        )

        if config['auths'][url] != encoded_credentials:
            config['auths'][url] = encoded_credentials
            self.write_config(path, config)

    @staticmethod
    def write_config(path, config):
        try:
            json.dump(config, open(path, "w"), indent=5, sort_keys=True)
        except Exception as exc:
            raise AnsibleContainerDockerConfigFileException("Failed to write docker registry config to %s - %s" %
                                                            (path, str(exc)))

    def push_latest_image(self, host, url=None, namespace=None):
        '''
        :param host: The host in the container.yml to push
        :parm url: URL of the registry to which images will be pushed
        :param namespace: namespace to append to the URL
        :return: None
        '''
        client = self.get_client()
        image_id, image_buildstamp = get_latest_image_for(self.project_name,
                                                          host, client)

        repository = "%s/%s-%s" % (namespace, self.project_name, host)
        if url != self.default_registry_url:
            url = REMOVE_HTTP.sub('', url)
            repository = "%s/%s" % (re.sub('/$', '', url), repository)

        logger.info('Tagging %s' % repository)
        client.tag(image_id, repository, tag=image_buildstamp)

        logger.info('Pushing %s:%s...' % (repository, image_buildstamp))
        stream = client.push(repository,
                             tag=image_buildstamp,
                             stream=True)
        last_status = None
        for data in stream:
            data = data.splitlines()
            for line in data:
                line = json.loads(line)
                if type(line) is dict and 'error' in line:
                    logger.error(line['error'])
                if type(line) is dict and 'status' in line:
                    if line['status'] != last_status:
                        logger.info(line['status'])
                    last_status = line['status']
                else:
                    logger.debug(line)

    def get_client(self):
        if not self._client:
            # To ensure version compatibility, we have to generate the kwargs ourselves
            client_kwargs = kwargs_from_env(assert_hostname=False)
            self._client = docker.AutoVersionClient(**client_kwargs)
            self.api_version = self._client.version()['ApiVersion']
            # Set the version in the env so it can be used elsewhere
            os.environ['DOCKER_API_VERSION'] = self.api_version
        return self._client

    def print_version_info(self):
        client = self.get_client()
        pprint.pprint(client.info())
        pprint.pprint(client.version())

    def bootstrap_env(self, temp_dir, behavior, operation, compose_option,
                      builder_img_id=None, context=None):
        """
        Build common Docker Compose elements required to execute orchestrate,
        terminate, restart, etc.
        
        :param temp_dir: A temporary directory usable as workspace
        :param behavior: x in x_operation_extra_args
        :param operation: Operation to perform, like, build, run, listhosts, etc
        :param compose_option: x in DEFAULT_COMPOSE_X_OPTIONS
        :param builder_img_id: Ansible Container Builder Image ID
        :param context: extra context to send to jinja_render_to_temp
        :return: options (options to pass to compose),
                 command_options (operation options to pass to compose),
                 command (compose's top level command)
        """

        if context is None:
            context = {}

        self.temp_dir = temp_dir
        extra_options = getattr(self, '{}_{}_extra_args'.format(behavior,
                                                                operation))()
        config = getattr(self, 'get_config_for_%s' % operation)()
        logger.debug('%s' % (config,))
        config_yaml = yaml_dump(config) if config else ''
        logger.debug('Config YAML is')
        logger.debug(config_yaml)
        hosts = self.all_hosts_in_orchestration()
        if operation == 'build' and self.params.get('service'):
            # build operation is limited to a specific list of services
            hosts = list(set(hosts).intersection(self.params['service']))
        jinja_render_to_temp('%s-docker-compose.j2.yml' % (operation,),
                             temp_dir,
                             'docker-compose.yml',
                             hosts=hosts,
                             project_name=self.project_name,
                             base_path=self.base_path,
                             params=self.params,
                             api_version=self.api_version,
                             builder_img_id=builder_img_id,
                             config=config_yaml,
                             env=os.environ,
                             **context)
        options = self.DEFAULT_COMPOSE_OPTIONS.copy()

        options.update({
            u'--verbose': self.params['debug'],
            u'--file': [
                os.path.join(temp_dir,
                             'docker-compose.yml')],
            u'--project-name': 'ansible',
        })
        command_options = getattr(self, 'DEFAULT_COMPOSE_{}_OPTIONS'.format(
            compose_option.upper())).copy()
        command_options.update(extra_options)

        project = project_from_options(self.base_path + '/ansible', options)
        command = main.TopLevelCommand(project)

        return options, command_options, command



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

class AnsibleContainerNotInitializedException(Exception):
    pass


class AnsibleContainerAlreadyInitializedException(Exception):
    pass

class AnsibleContainerNoAuthenticationProvidedException(Exception):
    pass

class AnsibleContainrRolesPathCreationException(Exception):
    pass

class AnsibleContainerDockerConfigFileException(Exception):
    pass

class AnsibleContainerDockerLoginException(Exception):
    pass

class AnsibleContainerConfigException(Exception):
    pass

class AnsibleContainerRegistryNotFoundException(Exception):
    pass

class AnsibleContainerRegistryAttributeException(Exception):
    pass

class AnsibleContainerMissingRegistryName(Exception):
    pass

class AnsibleContainerNoMatchingHosts(Exception):
    pass

class AnsibleContainerHostNotTouchedByPlaybook(Exception):
    pass

class AnsibleContainerShipItException(Exception):

    def __init__(self, msg, stdout=None, stderr=None):
        self.stderr = stderr
        self.stdout = stdout

        Exception.__init__(self, msg)

class AnsibleContainerFilterException(Exception):
    pass




# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

from . import LookupBase
from container import __version__

class LookupModule(LookupBase):

    def run(self, terms, variables, **kwargs):
        '''
        Return the Ansible Container version. Probably not very useful. Use to
        test that local filters are working.

        :param terms:
        :param variables:
        :param kwargs:
        :return: list
        '''
        return [__version__]


# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
import os.path
import json

from ..base_engine import BaseShipItEngine
from .deployment import Deployment
from .route import Route
from .service import Service
from ..utils import create_path
from ..constants import SHIPIT_PATH, SHIPIT_CONFIG_PATH

logger = logging.getLogger(__name__)


class ShipItEngine(BaseShipItEngine):
    name = u'openshift'

    def add_options(self, subparser):
        super(ShipItEngine, self).add_options(subparser)

    def run(self):
        tasks = []
        tasks += Service(config=self.config, project_name=self.project_name).get_task()
        tasks += Route(config=self.config, project_name=self.project_name).get_task()
        tasks += Deployment(config=self.config, project_name=self.project_name).get_task()
        self.init_role()
        self.create_role(tasks)
        self.create_playbook()

    def save_config(self):
        dest_path = os.path.join(self.base_path, SHIPIT_PATH, SHIPIT_CONFIG_PATH, self.name)
        create_path(dest_path)

        templates = Service(config=self.config, project_name=self.project_name).get_template()
        for template in templates:
            name = "%s-service.json" % template['metadata']['name']
            with open(os.path.join(dest_path, name), 'w') as f:
                f.write(json.dumps(template, indent=4))

        templates = Route(config=self.config, project_name=self.project_name).get_template()
        for template in templates:
            name = "%s.json" % template['metadata']['name']
            with open(os.path.join(dest_path, name), 'w') as f:
                f.write(json.dumps(template, indent=4))

        templates = Deployment(config=self.config, project_name=self.project_name).get_template()
        for template in templates:
            name = "%s-deployment.json" % template['metadata']['name']
            with open(os.path.join(dest_path, name), 'w') as f:
                f.write(json.dumps(template, indent=4))

        return dest_path







# -*- coding: utf-8 -*-

from __future__ import absolute_import


import logging
import re
import shlex

from collections import OrderedDict

logger = logging.getLogger(__name__)

DOCKER_VOL_PERMISSIONS = ['rw', 'ro', 'z', 'Z']


class Deployment(object):

    def __init__(self, config=None, project_name=None):
        self.project_name = project_name
        self.config = config

    def get_template(self, service_names=None):
        return self._get_template_or_task(request_type="config", service_names=service_names)

    def get_task(self, service_names=None):
        return self._get_template_or_task(request_type="task", service_names=service_names)

    def _get_template_or_task(self, request_type="task", service_names=None):
        templates = []
        for name, service in self.config.get('services', {}).items():
            new_template = self._create(request_type, name, service)
            if new_template:
                templates.append(new_template)
        return templates

    def _create(self, type, name, service):
        '''
        Creates a deployment template or playbook task
        '''

        container, volumes, pod = self._service_to_container(name, service, type=type)
        labels = dict(
            app=self.project_name,
            service=name
        )

        if type == 'config':
            state = 'present'
            if pod.get('state'):
                state = pod.pop('state')

            if state != 'absent':
                template = dict(
                    apiVersion="extensions/v1beta1",
                    kind="Deployment",
                    metadata=dict(
                        name=name,
                        labels=labels
                    ),
                    spec=dict(
                        template=dict(
                            metadata=dict(
                                labels=labels.copy()
                            ),
                            spec=dict(
                                containers=[container],
                            )
                        ),
                        replicas=1,
                        strategy=dict(
                            type='RollingUpdate'
                        )
                    )
                )
                if volumes:
                    template['spec']['template']['spec']['volumes'] = volumes
                if pod:
                    for key, value in pod.items():
                        if key == 'replicas':
                            template['spec'][key] = value
        else:
            template = dict(
                kube_deployment=OrderedDict(
                    deployment_name=name,
                    labels=labels.copy(),
                    replace=True,
                    containers=[container],
                )
            )
            if volumes:
                template['kube_deployment']['volumes'] = volumes
            if pod:
                template['kube_deployment'].update(pod)

        return template

    def _service_to_container(self, name, service, type="task"):
        '''
        Turn a service into a container and set of volumes. Maps Docker run directives
        to the Kubernetes container spec: http://kubernetes.io/docs/api-reference/v1/definitions/#_v1_container

        :param name:str: Name of the service
        :param service:dict: Configuration
        :param type: task or config
        :return: (container, volumes, pod)
        '''

        container = OrderedDict(
            name=name,
            securityContext=dict()
        )
        volumes = []
        pod = {}

        IGNORE_DIRECTIVES = [
            'aliases',
            'build',
            'expose',
            'labels',
            'links',
            'cgroup_parent',
            'dev_options',
            'devices',
            'depends_on',
            'dns',
            'dns_search',
            'env_file',        # TODO: build support for this?
            'user',            # needs to map to securityContext.runAsUser, which requires a UID
            'extends',
            'extrenal_links',
            'extra_hosts',
            'ipv4_address',
            'ipv6_address'
            'labels',
            'links',           # TODO: Add env vars?
            'logging',
            'log_driver',
            'lop_opt',
            'net',
            'network_mode',
            'networks',
            'restart',         # for replication controller, should be Always
            'pid',             # could map to pod.hostPID
            'security_opt',
            'stop_signal',
            'ulimits',
            'cpu_shares',
            'cpu_quota',
            'cpuset',
            'domainname',
            'hostname',
            'ipc',
            'mac_address',
            'mem_limit',
            'memswap_limit',
            'shm_size',
            'tmpfs',
            'options',
            'volume_driver',
            'volumes_from',   #TODO: figure out how to map?
        ]

        DOCKER_TO_KUBE_CAPABILITY_MAPPING=dict(
            SETPCAP='CAP_SETPCAP',
            SYS_MODULE='CAP_SYS_MODULE',
            SYS_RAWIO='CAP_SYS_RAWIO',
            SYS_PACCT='CAP_SYS_PACCT',
            SYS_ADMIN='CAP_SYS_ADMIN',
            SYS_NICE='CAP_SYS_NICE',
            SYS_RESOURCE='CAP_SYS_RESOURCE',
            SYS_TIME='CAP_SYS_TIME',
            SYS_TTY_CONFIG='CAP_SYS_TTY_CONFIG',
            MKNOD='CAP_MKNOD',
            AUDIT_WRITE='CAP_AUDIT_WRITE',
            AUDIT_CONTROL='CAP_AUDIT_CONTROL',
            MAC_OVERRIDE='CAP_MAC_OVERRIDE',
            MAC_ADMIN='CAP_MAC_ADMIN',
            NET_ADMIN='CAP_NET_ADMIN',
            SYSLOG='CAP_SYSLOG',
            CHOWN='CAP_CHOWN',
            NET_RAW='CAP_NET_RAW',
            DAC_OVERRIDE='CAP_DAC_OVERRIDE',
            FOWNER='CAP_FOWNER',
            DAC_READ_SEARCH='CAP_DAC_READ_SEARCH',
            FSETID='CAP_FSETID',
            KILL='CAP_KILL',
            SETGID='CAP_SETGID',
            SETUID='CAP_SETUID',
            LINUX_IMMUTABLE='CAP_LINUX_IMMUTABLE',
            NET_BIND_SERVICE='CAP_NET_BIND_SERVICE',
            NET_BROADCAST='CAP_NET_BROADCAST',
            IPC_LOCK='CAP_IPC_LOCK',
            IPC_OWNER='CAP_IPC_OWNER',
            SYS_CHROOT='CAP_SYS_CHROOT',
            SYS_PTRACE='CAP_SYS_PTRACE',
            SYS_BOOT='CAP_SYS_BOOT',
            LEASE='CAP_LEASE',
            SETFCAP='CAP_SETFCAP',
            WAKE_ALARM='CAP_WAKE_ALARM',
            BLOCK_SUSPEND='CAP_BLOCK_SUSPEND'
        )

        for key, value in service.items():
            if key in IGNORE_DIRECTIVES:
                pass
            elif key == 'cap_add':
                if not container['securityContext'].get('Capabilities'):
                    container['securityContext']['Capabilities'] = dict(add=[], drop=[])
                for cap in value:
                    if DOCKER_TO_KUBE_CAPABILITY_MAPPING[cap]:
                        container['securityContext']['Capabilities']['add'].append(DOCKER_TO_KUBE_CAPABILITY_MAPPING[cap])
            elif key == 'cap_drop':
                if not container['securityContext'].get('Capabilities'):
                    container['securityContext']['Capabilities'] = dict(add=[], drop=[])
                for cap in value:
                    if DOCKER_TO_KUBE_CAPABILITY_MAPPING[cap]:
                        container['securityContext']['Capabilities']['drop'].append(DOCKER_TO_KUBE_CAPABILITY_MAPPING[cap])
            elif key == 'command':
                if isinstance(value, basestring):
                    container['args'] = shlex.split(value)
                else:
                    container['args'] = value
            elif key == 'container_name':
                    container['name'] = value
            elif key == 'entrypoint':
                if isinstance(value, basestring):
                    container['command'] = shlex.split(value)
                else:
                    container['command'] = value
            elif key == 'environment':
                expanded_vars = self._expand_env_vars(value)
                if type == 'config':
                    container['env'] = expanded_vars
                else:
                    container['env'] = self._env_vars_to_task(expanded_vars)
            elif key in ('ports', 'expose'):
                if not container.get('ports'):
                    container['ports'] = []
                self._get_ports(value, type, container['ports'])
            elif key == 'privileged':
                container['securityContext']['privileged'] = value
            elif key == 'read_only':
                container['securityContext']['readOnlyRootFileSystem'] = value
            elif key == 'stdin_open':
                container['stdin'] = value
            elif key == 'volumes':
                vols, vol_mounts = self._kube_volumes(value)
                if vol_mounts:
                    container['volumeMounts'] = vol_mounts
                if vols:
                    volumes += vols
            elif key == 'working_dir':
                container['workingDir'] = value
            else:
                container[key] = value

        # Translate options:
        if service.get('options', {}).get('kube'):
            for key, value in service['options']['kube'].items():
                if key == 'seLinuxOptions':
                    container['securityContext']['seLinuxOptions'] = value
                elif key == 'runAsNonRoot':
                    container['securityContext']['runAsNonRoot'] = value
                elif key == 'runAsUser':
                    container['securityContext']['runAsUser'] = value
                elif key == 'replicas':
                    pod['replicas'] = value
                elif key == 'state':
                    pod['state'] = value

        return container, volumes, pod

    def _kube_volumes(self, docker_volumes):
        '''
        Given an array of Docker volumes return a set of volumes and a set of volumeMounts

        :param volumes: array of Docker volumes
        :return: (volumes, volumeMounts) - where each is a list of dicts
        '''
        volumes = []
        volume_mounts = []
        for vol in docker_volumes:
            logger.debug("volume: %s" % vol)
            source = None
            destination = None
            permissions = None
            if ':' in vol:
                pieces = vol.split(':')
                if len(pieces) == 3:
                    source, destination, permissions = vol.split(':')
                elif len(pieces) == 2:
                    if pieces[1] in DOCKER_VOL_PERMISSIONS:
                        destination, permissions = vol.split(':')
                    else:
                        source, destination = vol.split(':')
            else:
                destination = vol
            logger.debug("source: %s destination: %s permissions: %s" % (source, destination, permissions))
            named = False
            if destination:
                # slugify the destination to create a name
                name = re.sub(r'\/', '-', destination)
                name = re.sub(r'-', '', name, 1)

            if source:
                if re.match(r'[~./]', source):
                    # Source is a host path. We'll assume it exists on the host machine?
                    volumes.append(dict(
                        name=name,
                        hostPath=dict(
                            path=source
                        )
                    ))
                else:
                    # Named volume. The volume should be defined elsewhere.
                    name = source
                    named = True
            else:
                # Volume with no source, a.k.a emptyDir
                volumes.append(dict(
                    name=name,
                    emptyDir=dict(
                        medium=""
                    ),
                ))

            if not named:
                volume_mounts.append(dict(
                    mountPath=destination,
                    name=name,
                    readOnly=(True if permissions == 'ro' else False)
                ))

        return volumes, volume_mounts

    def _get_ports(self, ports, type, existing_ports):
        '''
        Determine the list of ports to expose from the container, and add to existing ports.

        :param: ports: list of port mappings
        :param: type: 'config' or 'task'
        :param: existing_ports: list of existing or already discovered ports
        :return: None
        '''
        for port in ports:
            if ':' in port:
                parts = port.split(':')
                if not self._port_exists(parts[1], existing_ports):
                    if type == 'config':
                        existing_ports.append(dict(containerPort=int(parts[1])))
                    else:
                        existing_ports.append(int(parts[1]))
            else:
                if not self._port_exists(port, existing_ports):
                    if type == 'config':
                        existing_ports.append(dict(containerPort=int(port)))
                    else:
                        existing_ports.append(int(port))

    @staticmethod
    def _port_exists(port, ports):
        found = False
        for p in ports:
            if p['containerPort'] == int(port):
                found = True
                break
        return found

    @staticmethod
    def _env_vars_to_task(env_vars):
        '''
        Turn list of vars into a dict for playbook task.
        :param env_variables: list of dicts
        :return: dict
        '''
        result = dict()
        for var in env_vars:
            result[var['name']] = var['value']
        return result

    def _expand_env_vars(self, env_variables):
        '''
        Turn containier environment attribute into dictionary of name/value pairs.

        :param env_variables: container env attribute value
        :type env_variables: dict or list
        :return: dict
        '''
        def r(x, y):
            if re.match('shipit_', x, flags=re.I):
                return dict(name=re.sub('^shipit_', '', x, flags=re.I), value=self._resolve_resource(y))
            return dict(name=x, value=y)

        results = []
        if isinstance(env_variables, dict):
            for key, value in env_variables.items():
                results.append(r(key, value))
        elif isinstance(env_variables, list):
            for envvar in env_variables:
                parts = envvar.split('=')
                if len(parts) == 1:
                    results.append(dict(name=re.sub('^shipit_', '', parts[0], flags=re.I), value=None))
                elif len(parts) == 2:
                    results.append(r(parts[0], parts[1]))
        return results

    def _resolve_resource(self, path):
        result = path
        if '/' in path:
            # TODO - support other resource types?
            res_type, res_name = path.split('/')
            if res_type == 'service':
                parts = res_name.split(':')
                result = unicode("{{ %s_service.spec.clusterIP }}:%s" % (parts[0].replace('-', '_'), parts[1]))
        return result



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import glob
import os
import imp

from container.exceptions import AnsibleContainerFilterException


def get_package_path(package):
    m = __import__(package)
    parts = package.split('.')[1:]
    for part in parts:
        m = getattr(m, part)
    package_path = os.path.dirname(m.__file__)
    return package_path


def get_filters(package_path, local=False):
    matches = (glob.glob(os.path.join(package_path, "*.py")))
    for path in matches:
        name, _ = os.path.splitext(path)
        if '__init__' in name:
            continue
        if not local:
            with open(path, 'r') as module_file:
                module_src = imp.load_source(name, path, module_file)
        else:
            module_src = __import__('container.filters.{0}'.format(os.path.basename(name)),
                                    fromlist='FilterModule')
        yield getattr(module_src, 'FilterModule')()


ANSIBLE_LOOKUP_NAME = 'ansible.plugins.lookup'
ANSIBLE_FILTERS_NAME = 'ansible.plugins.filter'


class LookupLoader(object):

    module_cache = {}

    def get(self, name):
        '''
        Load a lookup class and return an instance.

        :param name: string containing the name of the lookup submodule to load.
        :return: instance of the LookupModule class defined in the submodule
        '''
        obj = None
        package_path = os.path.dirname(__file__)
        path = os.path.join(package_path, 'lookup', "{0}.py".format(name.lower()))
        if os.path.isfile(path):
            try:
                if not self.module_cache.get(path):
                    self.module_cache[path] = __import__('container.filters.lookup.{0}'.format(name),
                                                         fromlist='LookupModule')
                obj = getattr(self.module_cache[path], 'LookupModule')()
            except Exception as exc:
                logger.debug("Unable to find local filter {0} - {1}".format(name, str(exc)))
                pass

        if not obj:
            # Attempt to load Ansible filter
            package_path = get_package_path(ANSIBLE_LOOKUP_NAME)
            path = os.path.join(package_path, "{0}.py".format(name.lower()))
            if os.path.isfile(path):
                try:
                    if not self.module_cache.get(path):
                        name, _ = os.path.splitext(path)
                        with open(path, 'r') as module_file:
                            self.module_cache[path] = imp.load_source(name, path, module_file)
                    obj = getattr(self.module_cache[path], 'LookupModule')()
                except Exception as exc:
                    raise AnsibleContainerFilterException("Failed to load lookup filter {0} - {1}".format(name,
                                                                                                          str(exc)))
            else:
                raise AnsibleContainerFilterException("Filter {0} not found.".format(name))
        return obj


class FilterLoader(object):

    all_filters = {}

    def all(self):
        '''
        Load all Jinja filters found in Ansible and locally with local filters taking precedence.

        :return: dict
        '''
        if not self.all_filters:
            try:
                package_path = get_package_path(ANSIBLE_FILTERS_NAME)
                for obj in get_filters(package_path):
                    self.all_filters.update(obj.filters())
            except Exception as exc:
                logger.debug('Failed to load ansible.plugin.filters - {0}'.format(str(exc)))
                pass

            try:
                package_path = os.path.dirname(__file__)
                for obj in get_filters(package_path, local=True):
                    self.all_filters.update(obj.filters())
            except Exception as exc:
                logger.debug('Failed to load plugin.filter - {0}'.format(str(exc)))
                pass

        return self.all_filters


class FilterBase(object):

    def filters(self):
        '''
        Get available filters.
        :return: dict of filters
        '''
        raise NotImplementedError()






#!/usr/bin/python
#
# Copyright 2016 Ansible by Red Hat
#
# This file is part of ansible-container
#

DOCUMENTATION = '''

module: oso_deployment

short_description: Start, cancel or retry a deployment on a Kubernetes or OpenShift cluster.

description:
  - Start, cancel or retry a deployment on a Kubernetes or OpenShift cluster by setting the C(state) to I(present) or
    I(absent).
  - Supports check mode. Use check mode to view a list of actions the module will take.

options:

'''

EXAMPLES = '''
'''

RETURN = '''
'''
import logging
import logging.config

from ansible.module_utils.basic import *

logger = logging.getLogger('oso_deployment')

LOGGING = (
    {
        'version': 1,
        'disable_existing_loggers': True,
        'handlers': {
            'null': {
                'level': 'DEBUG',
                'class': 'logging.NullHandler',
            },
            'file': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': 'ansible-container.log'
            }
        },
        'loggers': {
            'oso_deployment': {
                'handlers': ['file'],
                'level': 'INFO',
            },
            'container': {
                'handlers': ['file'],
                'level': 'INFO',
            },
            'compose': {
                'handlers': [],
                'level': 'INFO'
            },
            'docker': {
                'handlers': [],
                'level': 'INFO'
            }
        },
    }
)


class DeploymentManager(object):

    def __init__(self):

        self.arg_spec = dict(
            project_name=dict(type='str', aliases=['namespace'], required=True),
            state=dict(type='str', choices=['present', 'absent'], default='present'),
            labels=dict(type='dict'),
            deployment_name=dict(type='str', aliases=['name']),
            recreate=dict(type='bool', default=False),
            replace=dict(type='bool', default=True),
            replicas=dict(type='int', default=1),
            containers=dict(type='list'),
            strategy=dict(type='str', default='Rolling', choices=['Recreate', 'Rolling']),
            volumes=dict(type='list'),
        )

        self.module = AnsibleModule(self.arg_spec,
                                    supports_check_mode=True)

        self.project_name = None
        self.state = None
        self.labels = None
        self.deployment_name = None
        self.replace = None
        self.replicas = None
        self.containers = None
        self.strategy = None
        self.recreate = None
        self.volumes = None
        self.api = None
        self.check_mode = self.module.check_mode
        self.debug = self.module._debug

    def exec_module(self):

        for key in self.arg_spec:
            setattr(self, key, self.module.params.get(key))

        if self.debug:
            LOGGING['loggers']['container']['level'] = 'DEBUG'
            LOGGING['loggers']['oso_deployment']['level'] = 'DEBUG'
        logging.config.dictConfig(LOGGING)

        self.api = OriginAPI(self.module)

        actions = []
        changed = False
        deployments = dict()
        results = dict()

        project_switch = self.api.set_project(self.project_name)
        if not project_switch:
            actions.append("Create project %s" % self.project_name)
            if not self.check_mode:
                self.api.create_project(self.project_name)

        if self.state == 'present':
            deployment = self.api.get_resource('dc', self.deployment_name)
            if not deployment:
                template = self._create_template()
                changed = True
                actions.append("Create deployment %s" % self.deployment_name)
                if not self.check_mode:
                    self.api.create_from_template(template=template)
            elif deployment and self.recreate:
                actions.append("Delete deployment %s" % self.deployment_name)
                changed = True
                template = self._create_template()
                if not self.check_mode:
                    self.api.delete_resource('dc', self.deployment_name)
                    self.api.create_from_template(template=template)
            elif deployment and self.replace:
                template = self._create_template()
                template['status'] = dict(latestVersion=deployment['status']['latestVersion'] + 1)
                changed = True
                actions.append("Update deployment %s" % self.deployment_name)
                if not self.check_mode:
                    self.api.replace_from_template(template=template)

            deployments[self.deployment_name.replace('-', '_') + '_deployment'] = self.api.get_resource('dc', self.deployment_name)

        elif self.state == 'absent':
            if self.api.get_resource('dc', self.deployment_name):
                changed = True
                actions.append("Delete deployment %s" % self.deployment_name)
                if not self.check_mode:
                    self.api.delete_resource('dc', self.deployment_name)

        results['changed'] = changed
        if self.check_mode:
            results['actions'] = actions

        if deployments:
            results['ansible_facts'] = deployments

        self.module.exit_json(**results)

    def _create_template(self):

        for container in self.containers:
            if container.get('env'):
                container['env'] = self._env_to_list(container['env'])
            if container.get('ports'):
                container['ports'] = self._port_to_container_ports(container['ports'])

        template = dict(
            apiVersion="v1",
            kind="DeploymentConfig",
            metadata=dict(
                name=self.deployment_name,
            ),
            spec=dict(
                template=dict(
                    metadata=dict(),
                    spec=dict(
                        containers=self.containers
                    )
                ),
                replicas=self.replicas,
                strategy=dict(
                    type=self.strategy,
                ),
            )
        )

        if self.volumes:
            template['spec']['template']['spec']['volumes'] = self.volumes

        if self.labels:
            template['metadata']['labels'] = self.labels
            template['spec']['template']['metadata']['labels'] = self.labels

        return template

    @staticmethod
    def _env_to_list(env_variables):
        result = []
        for name, value in env_variables.items():
            result.append(dict(
                name=name,
                value=value
            ))
        return result

    @staticmethod
    def _port_to_container_ports(ports):
        result = []
        for port in ports:
            result.append(dict(containerPort=port))
        return result

#The following will be included by `ansble-container shipit` when cloud modules are copied into the role library path.
#include--> oso_api.py


def main():
    manager = DeploymentManager()
    manager.exec_module()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import os
import datetime
import re
import sys
import urllib
import gzip
import tarfile
import StringIO

import requests

from .exceptions import AnsibleContainerAlreadyInitializedException, \
                        AnsibleContainerRegistryAttributeException, \
                        AnsibleContainerHostNotTouchedByPlaybook
from .utils import *
from . import __version__

REMOVE_HTTP = re.compile('^https?://')

class BaseEngine(object):
    engine_name = None
    orchestrator_name = None
    default_registry_url = ''
    default_registry_name = ''

    def __init__(self, base_path, project_name, params={}):
        self.base_path = base_path
        self.project_name = project_name
        self.var_file = params.get('var_file')
        self.config = get_config(base_path, var_file=self.var_file)
        self.params = params
        self.support_init = True
        self.supports_build = True
        self.supports_push = True
        self.supports_run = True

        logger.debug('Initialized with params: %s', params)


    def all_hosts_in_orchestration(self):
        """
        List all hosts being orchestrated by the compose engine.

        :return: list of strings
        """
        raise NotImplementedError()

    def hosts_touched_by_playbook(self):
        """
        List all hosts touched by the execution of the build playbook.

        :return: list of strings
        """
        raise NotImplementedError()

    def build_buildcontainer_image(self):
        """
        Build in the container engine the builder container

        :return: generator of strings
        """
        raise NotImplementedError()

    def get_image_id_by_tag(self, name):
        """
        Query the engine to get an image identifier by tag

        :param name: the image name
        :return: the image identifier
        """
        raise NotImplementedError()

    def get_container_id_by_name(self, name):
        """
        Query the engine to get a container identifier by name

        :param name: the container name
        :return: the container identifier
        """
        raise NotImplementedError()

    def remove_container_by_name(self, name):
        """
        Remove a container from the engine given its name

        :param name: the name of the container to remove
        :return: None
        """
        raise NotImplementedError()

    def remove_container_by_id(self, id):
        """
        Remove a container from the engine given its identifier

        :param id: container identifier
        :return: None
        """
        raise NotImplementedError()

    def get_builder_image_id(self):
        """
        Query the enginer to get the builder image identifier

        :return: the image identifier
        """
        raise NotImplementedError()

    def get_builder_container_id(self):
        """
        Query the enginer to get the builder container identifier

        :return: the container identifier
        """
        raise NotImplementedError()

    def build_was_successful(self):
        """
        After the build completed, did the build run successfully?

        :return: bool
        """
        raise NotImplementedError()

    def orchestrate(self, operation, temp_dir, hosts=[]):
        """
        Execute the compose engine.

        :param operation: One of build, run, or listhosts
        :param temp_dir: A temporary directory usable as workspace
        :param hosts: (optional) A list of hosts to limit orchestration to
        :return: The exit status of the builder container (None if it wasn't run)
        """
        raise NotImplementedError()

    def orchestrate_build_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during build.

        :return: dictionary
        """
        raise NotImplementedError()

    def orchestrate_run_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during run.

        :return: dictionary
        """
        raise NotImplementedError()

    def orchestrate_install_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during install calls.

        :return: dictionary
        """
        return {}

    def orchestrate_listhosts_args(self):
        """
        Provide extra arguments to provide the orchestrator during listhosts.

        :return: dictionary
        """
        raise NotImplementedError()

    def terminate(self, operation, temp_dir, hosts=[]):
        """
        Stop, remove containers deployed by `orchestrate`.

        :return: dictionary
        """
        raise NotImplementedError()

    def terminate_stop_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during stop.

        :return: dictionary
        """
        return NotImplementedError

    def restart(self, operation, temp_dir, hosts=[]):
        """
        Restart containers deployed by `orchestrate`, deploys if not deployed

        :param operation: 'restart'
        :param temp_dir: A temporary directory usable as workspace
        :param hosts: (optional) A list of hosts to limit orchestration to
        """

        return NotImplementedError

    def restart_restart_extra_args(self):
        """
        Provide extra arguments to provide the orchestrator during restart.

        :return: dictionary
        """
        return NotImplementedError

    def post_build(self, host, version, flatten=True, purge_last=True):
        """
        After orchestrated build, prepare an image from the built container.

        :param host: the name of the host in the orchestration file
        :param version: the version tag for the resultant image
        :param flatten: whether to flatten the resulatant image all the way
        :param purge_last: whether to purge the last version of the image in the engine
        :return: None
        """
        raise NotImplementedError()

    def registry_login(self, username=None, password=None, email=None, url=None):
        """
        Logs into registry for this engine

        :param username: Username to login with
        :param password: Password to login with - None to prompt user
        :param email: Email address to login with
        :param url: URL of registry - default defined per backend
        :return: None
        """
        raise NotImplementedError()

    def currently_logged_in_registry_user(self, url):
        """
        Gets logged in user from configuration for a URL for the registry for
        this engine.

        :param url: URL
        :return: (username, email) tuple
        """
        raise NotImplementedError()

    def push_latest_image(self, host, url=None, namespace=None):
        """
        Push the latest built image for a host to a registry

        :param host: The host in the container.yml to push
        :param url: The url of the registry.
        :param namespace: The username or organization that owns the image repo
        :return: None
        """
        raise NotImplementedError()

    def get_config(self):
        raise NotImplementedError()

    def get_config_for_shipit(self, url=None, namespace=None):
        '''
        Get the configuration needed by cmdrun_shipit. Result should include
        the *options* attribute for each service, as it may contain cluster
        directives.

        :param url: URL to the registry.
        :param namepace: a namespace withing the registry. Typically a username or organization.
        :return: configuration dictionary
        '''
        raise NotImplementedError()

    def print_version_info(self):
        '''
        Output to stdout version information about this engine and orchestrator.

        :return: None
        '''


def cmdrun_init(base_path, project=None, **kwargs):
    if project:
        if os.listdir(base_path):
            raise AnsibleContainerAlreadyInitializedException(
                u'The init command can only be run in an empty directory.')
        try:
            namespace, name = project.split('.', 1)
        except ValueError:
            raise ValueError(u'Invalid project name: %r; use '
                             u'"username.project" style syntax.' % project)
        galaxy_base_url = kwargs.pop('server')
        response = requests.get(urllib.basejoin(galaxy_base_url,
                                                '/api/v1/roles/'),
                                params={'role_type': 'APP',
                                        'namespace': namespace,
                                        'name': name},
                                headers={'Accepts': 'application/json'})
        try:
            response.raise_for_status()
        except requests.RequestException, e:
            raise requests.RequestException(u'Could not find %r on Galaxy '
                                            u'server %r: %r' % (project,
                                                                galaxy_base_url,
                                                                e))
        if not response.json()['count']:
            raise ValueError(u'Could not find %r on Galaxy '
                             u'server %r: No such container app' % (project,
                                                                    galaxy_base_url))
        container_app_data = response.json()['results'][0]
        if not all([container_app_data[k] for k in ['github_user',
                                                    'github_repo']]):
            raise ValueError(u'Container app %r does not have a GitHub URL' % project)
        archive_url = u'https://github.com/%s/%s/archive/%s.tar.gz' % (
            container_app_data['github_user'],
            container_app_data['github_repo'],
            container_app_data['github_branch'] or u'master'
        )
        archive = requests.get(archive_url)
        try:
            archive.raise_for_status()
        except requests.RequestException, e:
            raise requests.RequestException(u'Could not get archive at '
                                            u'%r: %r' % (archive_url, e))
        faux_file = StringIO.StringIO(archive.content)
        gz_obj = gzip.GzipFile(fileobj=faux_file)
        tar_obj = tarfile.TarFile(fileobj=gz_obj)
        members = tar_obj.getmembers()
        # now we do the actual extraction to the path
        for member in members:
            # we only extract files, and remove any relative path
            # bits that might be in the file for security purposes
            # and drop the leading directory, as mentioned above
            if member.isreg() or member.issym():
                parts = member.name.split(os.sep)[1:]
                final_parts = []
                for part in parts:
                    if part != '..' and '~' not in part and '$' not in part:
                        final_parts.append(part)
                member.name = os.path.join(*final_parts)
                tar_obj.extract(member, base_path)
        logger.info(u'Ansible Container initialized from Galaxy container app %r', project)
    else:
        container_dir = os.path.normpath(
            os.path.join(base_path, 'ansible'))
        container_cfg = os.path.join(container_dir, 'container.yml')
        if os.path.exists(container_cfg):
            raise AnsibleContainerAlreadyInitializedException()
        if not os.path.exists(container_dir):
            os.mkdir(container_dir)
        template_dir = os.path.join(jinja_template_path(), 'ansible')
        context = {
            u'ansible_container_version': __version__
        }
        for tmpl_filename in os.listdir(template_dir):
            jinja_render_to_temp('ansible/%s' % tmpl_filename,
                                 container_dir,
                                 tmpl_filename.replace('.j2', ''),
                                 **context)
        logger.info('Ansible Container initialized.')


def cmdrun_build(base_path, engine_name, flatten=True, purge_last=True, local_builder=False,
                 rebuild=False, service=None, ansible_options='', save_build_container=False,
                 roles_path=None, **kwargs):
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)
    try:
        builder_img_id = engine_obj.get_image_id_by_tag(engine_obj.builder_container_img_tag)
    except NameError:
        if local_builder:
            create_build_container(engine_obj, base_path)
    with make_temp_dir() as temp_dir:
        logger.info('Starting %s engine to build your images...'
                    % engine_obj.orchestrator_name)
        touched_hosts = set(engine_obj.hosts_touched_by_playbook())
        if service:
            touched_hosts &= set(service)
            if not touched_hosts:
                raise AnsibleContainerHostNotTouchedByPlaybook()
        engine_obj.orchestrate('build', temp_dir, context=dict(rebuild=rebuild))
        if not engine_obj.build_was_successful():
            logger.error('Ansible playbook run failed.')
            if not save_build_container:
                logger.info('Cleaning up Ansible Container builder...')
                builder_container_id = engine_obj.get_builder_container_id()
                engine_obj.remove_container_by_id(builder_container_id)
            raise RuntimeError(u'Ansible build failed')
        # Cool - now export those containers as images
        version = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')
        logger.info('Exporting built containers as images...')
        for host in touched_hosts:
            engine_obj.post_build(host, version, flatten=flatten, purge_last=purge_last)
        if not save_build_container:
            logger.info('Cleaning up Ansible Container builder...')
            builder_container_id = engine_obj.get_builder_container_id()
            engine_obj.remove_container_by_id(builder_container_id)


def cmdrun_run(base_path, engine_name, service=[], production=False, **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)
    with make_temp_dir() as temp_dir:
        hosts = service or (engine_obj.all_hosts_in_orchestration())
        engine_obj.orchestrate('run', temp_dir,
                               hosts=hosts)


def cmdrun_stop(base_path, engine_name, service=[], **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)
    with make_temp_dir() as temp_dir:
        hosts = service or (engine_obj.all_hosts_in_orchestration())
        engine_obj.terminate('stop', temp_dir, hosts=hosts)


def cmdrun_restart(base_path, engine_name, service=[], **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)
    with make_temp_dir() as temp_dir:
        hosts = service or (engine_obj.all_hosts_in_orchestration())
        engine_obj.restart('restart', temp_dir, hosts=hosts)


def cmdrun_push(base_path, engine_name, username=None, password=None, email=None, push_to=None, **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)

    # resolve url and namespace
    config = engine_obj.config
    url = engine_obj.default_registry_url
    namespace = None
    if push_to:
        if config.get('registries', {}).get(push_to):
            url = config['registries'][push_to].get('url')
            namespace = config['registries'][push_to].get('namespace')
            if not url:
                raise AnsibleContainerRegistryAttributeException("Registry %s missing required attribute 'url'."
                                                                 % push_to)
        else:
            url, namespace = resolve_push_to(push_to, engine_obj.default_registry_url)

    # Check that we can authenticate to the registry and get the username
    username = engine_obj.registry_login(username=username, password=password,
                                         email=email, url=url)
    if not namespace:
        namespace = username

    logger.info('Pushing to "%s/%s' % (re.sub(r'/$', '', url), namespace))

    for host in engine_obj.hosts_touched_by_playbook():
        engine_obj.push_latest_image(host, url=url, namespace=namespace)
    logger.info('Done!')


def cmdrun_shipit(base_path, engine_name, pull_from=None, **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)
    shipit_engine_name = kwargs.pop('shipit_engine')
    project_name = os.path.basename(base_path).lower()

    # determine the registry url and namespace the cluster will use to pull images
    config = engine_obj.config
    url = None
    namespace = None
    if not pull_from:
        url = engine_obj.default_registry_url
    elif config.get('registries', {}).get(pull_from):
        url = config['registries'][pull_from].get('url')
        namespace = config['registries'][pull_from].get('namespace')
        if not url:
            raise AnsibleContainerRegistryAttributeException("Registry %s missing required attribute 'url'."
                                                             % pull_from)
        pull_from = None  # pull_from is now resolved to a url/namespace
    if url and not namespace:
        # try to get the username for the url from the container engine
        try:
            namespace = engine_obj.registry_login(url=url)
        except Exception as exc:
            if "Error while fetching server API version" in str(exc):
                msg = "Cannot connect to the Docker daemon. Is the daemon running?"
            else:
                msg = "Unable to determine namespace for registry %s. Error: %s. Either authenticate with the " \
                      "registry or provide a namespace for the registry in container.yml" % (url, str(exc))
            raise AnsibleContainerRegistryAttributeException(msg)

    config = engine_obj.get_config_for_shipit(pull_from=pull_from, url=url, namespace=namespace)

    shipit_engine_obj = load_shipit_engine(AVAILABLE_SHIPIT_ENGINES[shipit_engine_name]['cls'],
                                           config=config,
                                           base_path=base_path,
                                           project_name=project_name)

    # create the role and sample playbook
    shipit_engine_obj.run()
    logger.info('Role %s created.' % project_name)

    if kwargs.get('save_config'):
        # generate and save the configuration templates
        config_path = shipit_engine_obj.save_config()
        logger.info('Saved configuration to %s' % config_path)

def cmdrun_install(base_path, engine_name, roles=[], **kwargs):
    assert_initialized(base_path)
    engine_args = kwargs.copy()
    engine_args.update(locals())
    engine_obj = load_engine(**engine_args)

    with make_temp_dir() as temp_dir:
        engine_obj.orchestrate('install', temp_dir)


def cmdrun_version(base_path, engine_name, debug=False, **kwargs):
    print 'Ansible Container, version', __version__
    if debug:
        print u', '.join(os.uname())
        print sys.version, sys.executable
        assert_initialized(base_path)
        engine_args = kwargs.copy()
        engine_args.update(locals())
        engine_obj = load_engine(**engine_args)
        engine_obj.print_version_info()

def create_build_container(container_engine_obj, base_path):
    assert_initialized(base_path)
    logger.info('(Re)building the Ansible Container image.')
    build_output = container_engine_obj.build_buildcontainer_image()
    for line in build_output:
        logger.debug(line)
    builder_img_id = container_engine_obj.get_builder_image_id()
    logger.info('Ansible Container image has ID %s', builder_img_id)
    return builder_img_id


def resolve_push_to(push_to, default_url):
    '''
    Given a push-to value, return the registry and namespace.

    :param push_to: string: User supplied --push-to value.
    :param default_index: string: Container engine's default_index value (e.g. docker.io).
    :return: tuple: index_name, namespace
    '''
    protocol = 'http://' if push_to.startswith('http://') else 'https://'
    url = push_to = REMOVE_HTTP.sub('', push_to)
    namespace = None
    parts = url.split('/', 1)
    special_set = {'.', ':'}
    char_set = set([c for c in parts[0]])

    if len(parts) == 1:
        if not special_set.intersection(char_set) and parts[0] != 'localhost':
            registry_url = default_url
            namespace = push_to
        else:
            registry_url = protocol + parts[0]
    else:
        registry_url = protocol + parts[0]
        namespace = parts[1]

    return registry_url, namespace




# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)


class LookupBase(object):

    def run(self, terms, variables, **kwargs):
        raise NotImplementedError()
        return []




# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import os
import importlib
from jinja2 import Environment, FileSystemLoader

from .exceptions import AnsibleContainerNotInitializedException
from .config import AnsibleContainerConfig
from .temp import MakeTempDir


__all__ = ['make_temp_dir',
           'jinja_template_path',
           'jinja_render_to_temp',
           'get_config',
           'config_format_version',
           'assert_initialized',
           'get_latest_image_for',
           'load_engine',
           'load_shipit_engine',
           'AVAILABLE_SHIPIT_ENGINES']


AVAILABLE_SHIPIT_ENGINES = {
    'kube': {
        'help': 'Generate a role that deploys to Kubernetes.',
        'cls': 'kubernetes'
    },
    'openshift': {
        'help': 'Generate a role that deploys to OpenShift Origin.',
        'cls': 'openshift'
    }
}


make_temp_dir = MakeTempDir


def jinja_template_path():
    return os.path.normpath(
        os.path.join(
            os.path.dirname(__file__),
            'templates'))

def jinja_render_to_temp(template_file, temp_dir, dest_file, **context):
    j2_tmpl_path = jinja_template_path()
    j2_env = Environment(loader=FileSystemLoader(j2_tmpl_path))
    j2_tmpl = j2_env.get_template(template_file)
    rendered = j2_tmpl.render(dict(temp_dir=temp_dir, **context))
    logger.debug('Rendered Jinja Template:')
    logger.debug(rendered.encode('utf8'))
    open(os.path.join(temp_dir, dest_file), 'w').write(
        rendered.encode('utf8'))

def get_config(base_path, var_file=None):
    return AnsibleContainerConfig(base_path, var_file=var_file)

def config_format_version(base_path, config_data=None):
    if not config_data:
        config_data = get_config(base_path)
    return int(config_data.pop('version', 1))

def assert_initialized(base_path):
    ansible_dir = os.path.normpath(
        os.path.join(base_path, 'ansible'))
    container_file = os.path.join(ansible_dir, 'container.yml')
    ansible_file = os.path.join(ansible_dir, 'main.yml')
    if not os.path.exists(ansible_dir) or not os.path.isdir(ansible_dir) or \
            not os.path.exists(container_file) or not os.path.isfile(container_file) \
            or not os.path.exists(ansible_file) or not os.path.isfile(ansible_file):
        raise AnsibleContainerNotInitializedException()

def get_latest_image_for(project_name, host, client):
    image_data = client.images(
        '%s-%s' % (project_name, host,)
    )
    try:
        latest_image_data, = [datum for datum in image_data
                              if '%s-%s:latest' % (project_name, host,) in
                              datum['RepoTags']]
        image_buildstamp = [tag for tag in latest_image_data['RepoTags']
                            if not tag.endswith(':latest')][0].split(':')[-1]
        image_id = latest_image_data['Id']
        return image_id, image_buildstamp
    except (IndexError, ValueError):
        # No previous image built
        return None, None

def load_engine(engine_name='', base_path='', **kwargs):
    """

    :param engine_name: the string for the module containing the engine.py code
    :param base_path: the base path during operation
    :return: container.engine.BaseEngine
    """
    mod = importlib.import_module('container.%s.engine' % engine_name)
    project_name = os.path.basename(base_path).lower()
    logger.debug('Project name is %s', project_name)
    return mod.Engine(base_path, project_name, kwargs)


def load_shipit_engine(engine_class, **kwargs):
    '''
    Given a class name, dynamically load a shipit engine.

    :param engine_class: name of the shipit engine class
    :param kwargs: key/value args to pass to the new shipit engine obj.
    :return: shipit engine object
    '''
    try:
        engine_module = importlib.import_module(
            'container.shipit.%s.engine' % engine_class)
    except ImportError as exc:
        raise ImportError(
            'No shipit module for %s found - %s' % (engine_class, str(exc)))
    try:
        engine_cls = getattr(engine_module, 'ShipItEngine')
    except Exception as exc:
        raise ImportError('Error getting ShipItEngine for %s - %s' % (engine_class, str(exc)))

    return engine_cls(**kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import os
import json
import yaml
import re

from jinja2 import Environment, FileSystemLoader
from collections import Mapping
from .exceptions import AnsibleContainerConfigException
from .filters import LookupLoader, FilterLoader
from .temp import MakeTempDir as make_temp_dir

# TODO: Actually do some schema validation


class AnsibleContainerConfig(Mapping):
    _config = {}
    base_path = None
    lookup_loader = LookupLoader()
    filter_loader = FilterLoader()

    def __init__(self, base_path, var_file=None):
        self.base_path = base_path
        self.var_file = var_file
        self.config_path = os.path.join(self.base_path, 'ansible/container.yml')
        self.all_filters = self.filter_loader.all()
        self.set_env('prod')

    def set_env(self, env):
        '''
        Loads config from container.yml, performs Jinja templating, and stores the resulting dict to self._config.

        :param env: string of either 'dev' or 'prod'. Indicates 'dev_overrides' handling.
        :return: None
        '''
        assert env in ['dev', 'prod']
        context = self._get_variables()
        config = self._render_template(context=context)
        try:
            config = yaml.safe_load(config)
        except yaml.YAMLError as exc:
            raise AnsibleContainerConfigException(u"Parsing container.yml - %s" % str(exc))
        if config.get('defaults'):
            del config['defaults']

        for service, service_config in (config.get('services') or {}).items():
            if not service_config or isinstance(service_config, basestring):
                raise AnsibleContainerConfigException(u"Error: no definition found in container.yml for service %s."
                                                      % service)
            if isinstance(service_config, dict):
                dev_overrides = service_config.pop('dev_overrides', {})
                if env == 'dev':
                    service_config.update(dev_overrides)

        logger.debug(u"Config:\n%s" % json.dumps(config,
                                                 sort_keys=True,
                                                 indent=4,
                                                 separators=(',', ': ')))
        self._config = config

    def _lookup(self, name, *args, **kwargs):
        lookup_instance = self.lookup_loader.get(name)
        wantlist = kwargs.pop('wantlist', False)
        try:
            ran = lookup_instance.run(args, {}, **kwargs)
        except Exception as exc:
            raise AnsibleContainerConfigException("Error in filter %s - %s" % (name, exc))

        if ran and not wantlist:
            ran = ','.join(ran)
        return ran

    def _render_template(self, context=None, path=None, template='container.yml'):
        '''
        Apply Jinja template rendering to a given template. If no template provided, template ansible/container.yml

        :param template_vars: dict providing Jinja context
        :return: dict
        '''
        if not context:
            context = dict()
        if not path:
            path = os.path.join(self.base_path, 'ansible')
        j2_env = Environment(loader=FileSystemLoader(path))
        j2_env.globals['lookup'] = self._lookup
        j2_env.filters.update(self.all_filters)
        j2_tmpl = j2_env.get_template(template)
        return j2_tmpl.render(**context).encode('utf8')

    def _get_variables(self):
        '''
        Resolve variables by creating an empty dict and updating it first with the 'defaults' section in the config,
        then any variables from var_file, and finally any AC_* environment variables. Returns the resulting dict.

        :return: dict
        '''
        new_vars = {}
        new_vars.update(self._get_defaults())
        if self.var_file:
            logger.debug('Reading variables from var file...')
            file_vars = self._get_variables_from_file(self.var_file, context=new_vars)
            new_vars.update(file_vars)
        new_vars.update(self._get_environment_variables())
        logger.debug(u'Template variables:\n %s' % json.dumps(new_vars,
                                                              sort_keys=True,
                                                              indent=4,
                                                              separators=(',', ': ')))
        return new_vars

    def _get_defaults(self):
        '''
        Parse the optional 'defaults' section of container.yml

        :return: dict
        '''
        defaults = {}
        default_lines = ['defaults:']
        found = False
        sections = [u'version:', u'services:', u'registries:']
        try:
            with open(self.config_path, 'r') as f:
                for line in f:
                    if re.search(r'^defaults:', line):
                        found = True
                        continue
                    if found:
                        if re.sub(u'\n', '', line) not in sections:
                            default_lines.append(re.sub(u'\n', '', line))
                        else:
                            break
        except (OSError, IOError):
            raise AnsibleContainerConfigException(u"Failed to open %s. Are you in the correct directory?" %
                                                  self.config_path)

        if len(default_lines) > 1:
            # re-assemble the defaults section, template, and parse as yaml
            with make_temp_dir() as temp_dir:
                with open(os.path.join(temp_dir, 'defaults.txt'), 'w') as f:
                    f.write(u'\n'.join(default_lines))
                default_section = self._render_template(context={}, path=temp_dir, template='defaults.txt')
            try:
                config = yaml.safe_load(default_section)
                defaults.update(config.get('defaults'))
            except yaml.YAMLError as exc:
                raise AnsibleContainerConfigException(u"Parsing container.yml - %s" % str(exc))
        logger.debug(u"Default vars:")
        logger.debug(json.dumps(defaults, sort_keys=True, indent=4, separators=(',', ': ')))
        return defaults

    def _get_environment_variables(self):
        '''
        Look for any environment variables that start with 'AC_'. Returns dict of key:value pairs, where the
        key is the result of removing 'AC_' from the variable name and converting the remainder to lowercase.
        For example, 'AC_DEBUG=1' becomes 'debug: 1'.

        :return dict
        '''
        logger.debug(u'Getting environment variables...')
        new_vars = {}
        for var, value in os.environ.iteritems():
            matches = re.match(r'^AC_(.+)$', var)
            if matches:
                new_vars[matches.group(1).lower()] = value
        return new_vars

    def _get_variables_from_file(self, file, context=None):
        '''
        Looks for file relative to base_path. If not found, checks relative to base_path/ansible.
        If file extension is .yml | .yaml, parses as YAML, otherwise parses as JSON.

        :param file: string: path relative to base_path or base_path/ansible.
        :param context: dict of any available default variables
        :return: dict
        '''
        file_path = os.path.abspath(file)
        path = os.path.dirname(file_path)
        name = os.path.basename(file_path)
        if not os.path.isfile(file_path):
            path = self.base_path
            file_path = os.path.normpath(os.path.join(self.base_path, file))
            name = os.path.basename(file_path)
            if not os.path.isfile(file_path):
                path = os.path.join(self.base_path, 'ansible')
                file_path = os.path.normpath(os.path.join(path, file))
                name = os.path.basename(file_path)
                if not os.path.isfile(file_path):
                    raise AnsibleContainerConfigException(u"Unable to locate %s. Provide a path relative to %s or %s." % (
                                                          file, self.base_path, os.path.join(self.base_path, 'ansible')))
        logger.debug("Use variable file: %s" % file_path)
        data = self._render_template(context=context, path=path, template=name)

        if name.endswith('yml') or name.endswith('yaml'):
            try:
                config = yaml.safe_load(data)
            except yaml.YAMLError as exc:
                raise AnsibleContainerConfigException(u"YAML exception: %s" %  str(exc))
        else:
            try:
                config = json.loads(data)
            except Exception as exc:
                raise AnsibleContainerConfigException(u"JSON exception: %s" % str(exc))
        return config

    def __getitem__(self, item):
        return self._config.get(item)

    def __iter__(self):
        return iter(self._config)

    def __len__(self):
        return len(self._config)





#!/usr/bin/env python

import yaml
import json
import sys
import argparse
import re
import os


def config_keys():
    '''
    Get the list of hosts from env var ANSIBLE_ORCHESTRATED_HOSTS.

    :return: list
    '''
    return os.environ.get('ANSIBLE_ORCHESTRATED_HOSTS', '').split(',')


def cmd_list():
    hosts = config_keys()
    return dict(
        docker=hosts,
        _meta=dict(
            hostvars={
                host: {'ansible_host': 'ansible_%s_1' % host}
                for host in hosts
            }
        )
    )


def cmd_host(host):
    hosts = config_keys()
    if host not in hosts:
        return {}
    return dict(
        ansible_host='ansible_%s_1' % host
    )

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description=u'Dynamic inventory for ansible-container')
    parser.add_argument('--list', action='store_true', dest='list',
                        help=u'List hosts', default=False)
    parser.add_argument('--host', action='store', dest='host',
                        help=u'Host detail', default=None)
    args = parser.parse_args()
    if args.host:
        json.dump(cmd_host(args.host), sys.stdout)
    else:
        json.dump(cmd_list(), sys.stdout)




# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from . import FilterBase

logger = logging.getLogger(__name__)


def test_filter(*args, **kwargs):
    '''
    Not really a filter. Use to whether or not import method works.

    :param args:
    :param kwargs:
    :return:
    '''
    return 'success!'


class FilterModule(FilterBase):

    def filters(self):
        return {
            u'test_filter': test_filter
        }


from setuptools import setup, find_packages
from pip.req import parse_requirements
import container

install_reqs = parse_requirements('requirements.txt', session=False)
reqs = [str(ir.req) for ir in install_reqs]

setup(
    name='ansible-container',
    version=container.__version__,
    packages=find_packages(include='container.*'),
    include_package_data=True,
    url='https://github.com/ansible/ansible-container',
    license='LGPLv3 (See LICENSE file for terms)',
    author='Joshua "jag" Ginsberg, Chris Houseknecht, and others (See AUTHORS file for contributors)',
    author_email='jag@ansible.com',
    description=('Ansible Container empowers you to orchestrate, build, run, and ship '
                 'Docker images built from Ansible playbooks.'),
    entry_points={
        'console_scripts': ['ansible-container = container.cli:commandline']
    },
    install_requires=reqs
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

logger = logging.getLogger(__name__)

import tempfile
import shutil
import os
import sys

import ansible.constants as C
from ansible.galaxy import Galaxy
from ansible.galaxy.role import GalaxyRole
from ansible.playbook.role.requirement import RoleRequirement
import ruamel.yaml

ANSIBLE_CONTAINER_PATH = '/ansible-container'

class FatalException(Exception): pass
class RoleException(Exception): pass

class MakeTempDir(object):
    temp_dir = None

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()
        logger.debug('Using temporary directory %s...', self.temp_dir)
        return os.path.realpath(self.temp_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            logger.debug('Cleaning up temporary directory %s...', self.temp_dir)
            shutil.rmtree(self.temp_dir)
        except Exception, e:
            logger.exception('Failure cleaning up temp space')
            pass

class InCaseOfFail(object):
    def __init__(self, temp_dir):
        self.temp_dir = temp_dir

    def __enter__(self):
        for yml_file in ['container.yml', 'main.yml', 'requirements.yml']:
            shutil.copyfile(os.path.join(ANSIBLE_CONTAINER_PATH, 'ansible', yml_file),
                            os.path.join(self.temp_dir, yml_file))

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logger.info('Undoing changes to .yml files')
            for yml_file in ['container.yml', 'main.yml', 'requirements.yml']:
                shutil.copyfile(
                    os.path.join(self.temp_dir, yml_file),
                    os.path.join(ANSIBLE_CONTAINER_PATH, 'ansible', yml_file))

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

def get_galaxy(tmp_dir, token):
    return Galaxy(AttrDict(api_server=C.GALAXY_SERVER,
                           ignore_certs=C.GALAXY_IGNORE_CERTS,
                           ignore_errors=False,
                           no_deps=False,
                           roles_path=[tmp_dir],
                           token=token))

def role_to_temp_space(role_req, galaxy):
    role_req_kwargs = RoleRequirement.role_yaml_parse(role_req.strip())
    role_obj = GalaxyRole(galaxy, **role_req_kwargs)
    installed = role_obj.install()
    return role_obj, installed

def get_container_yml_snippet(role_obj):
    container_yml_path = os.path.join(role_obj.path, 'meta', 'container.yml')
    if os.path.exists(container_yml_path):
        try:
            snippet = ruamel.yaml.round_trip_load(open(container_yml_path))
        except Exception, e:
            logger.exception('Error loading container.yml snippet for %s',
                             role_obj)
            return None
        logger.debug('Role %s is containerized', role_obj)
        try:
            assert isinstance(snippet, dict) and len(snippet) == 1
        except AssertionError:
            logger.exception('Role %s container.yml is malformed', role_obj)
            return None
        return snippet
    logger.debug('No %s found for %s, not containerized', container_yml_path,
                 role_obj)

def get_knobs_and_dials(role_obj):
    defaults_yml_path = os.path.join(role_obj.path, 'defaults', 'main.yml')
    if os.path.exists(defaults_yml_path):
        try:
            defaults = ruamel.yaml.round_trip_load(open(defaults_yml_path))
        except Exception, e:
            logger.exception('Error loading defaults/main.yml for %s',
                             role_obj)
        else:
            return defaults
    return {}

def update_container_yml(role_obj):
    snippet = get_container_yml_snippet(role_obj)
    if not snippet:
        return None
    container_yml_path = os.path.join(ANSIBLE_CONTAINER_PATH, 'ansible',
                                      'container.yml')
    try:
        container_yml = ruamel.yaml.round_trip_load(open(container_yml_path))
    except Exception, e:
        logger.exception('Could not load project ansible/container.yml')
        raise FatalException()

    # It can be None if left empty
    if not container_yml['services']:
        container_yml['services'] = {}
    services = container_yml['services']
    # The snippet should be a dictionary with one key
    new_service_key = snippet.keys()[0]
    if new_service_key in services:
        logger.error('Role defines service %s, but ansible/container.yml '
                     'already has a service with this name', new_service_key)
        raise RoleException()

    services[new_service_key] = snippet[new_service_key]
    try:
        ruamel.yaml.round_trip_dump(container_yml,
                                    stream=open(container_yml_path, 'w'))
    except Exception, e:
        logger.exception('Error updating ansible/container.yml')
        raise FatalException()
    return new_service_key

def update_main_yml(service, role_obj):
    defaults = get_knobs_and_dials(role_obj)
    main_yml_path = os.path.join(ANSIBLE_CONTAINER_PATH, 'ansible',
                                 'main.yml')
    main_yml = None
    try:
        main_yml = ruamel.yaml.round_trip_load(open(main_yml_path))
    except Exception, e:
        logger.exception('Could not load project ansible/main.yml')
        raise FatalException()

    if not main_yml:
        main_yml = [] 

    # this is a ruamel.ordereddict.OrderedDict
    # for readability, put the role name at the start of the dict
    defaults.insert(0, 'role', role_obj.name)

    snippet = {
        'hosts': service,
        'roles': [defaults]
    }
    main_yml.append(snippet)
    try:
        ruamel.yaml.round_trip_dump(main_yml,
                                    stream=open(main_yml_path, 'w'))
    except Exception, e:
        logger.exception('Error updating ansible/main.yml')
        raise FatalException()

def update_requirements_yml(role_obj):
    requirements_yml_path = os.path.join(ANSIBLE_CONTAINER_PATH, 'ansible',
                                         'requirements.yml')
    requirements = None
    if os.path.exists(requirements_yml_path):
        try:
            requirements = ruamel.yaml.round_trip_load(open(requirements_yml_path)) or []
        except Exception, e:
            logger.exception('Could not load project ansible/requirements.yml')
            raise FatalException()
    if not requirements:
        requirements = []
    for req in requirements:
        if req.get('src', '') == role_obj.name:
            logger.warning('Requirement %s already found in requirements.yml',
                           role_obj.name)
            return
    requirements.append({'src': role_obj.name})
    try:
        ruamel.yaml.round_trip_dump(requirements,
                                    stream=open(requirements_yml_path, 'w'))
    except Exception, e:
        logger.exception('Error updating ansible/requirements.yml')
        raise FatalException()

def install(roles):
    roles_to_install = list(roles)
    with MakeTempDir() as temp_dir:
        galaxy = get_galaxy(temp_dir, None) # FIXME: support tokens
        roles_processed = []
        role_failure = False
        with InCaseOfFail(temp_dir):
            while roles_to_install:
                try:
                    role_to_install = roles_to_install.pop()
                    role_obj, installed = role_to_temp_space(role_to_install, galaxy)
                    if installed:
                        deps = role_obj.metadata.get('dependencies', [])
                        for dep in deps:
                            if dep not in roles_to_install + roles_processed:
                                roles_to_install.append(dep)
                        service = update_container_yml(role_obj)
                        if service:
                            update_main_yml(service, role_obj)
                        else:
                            logger.warning('Role %s is not Ansible Container enabled.',
                                           role_obj)
                        update_requirements_yml(role_obj)
                    roles_processed.append(role_to_install)
                except FatalException:
                    raise
                except RoleException:
                    role_failure = True
                    continue
    if role_failure:
        raise RoleException('One or more roles failed.')

if __name__ == '__main__':
    import argparse
    logging.basicConfig(level=logging.INFO, format='%(message)s')

    parser = argparse.ArgumentParser()
    parser.add_argument('roles', nargs='+', action='store')

    args = parser.parse_args()
    try:
        install(args.roles)
    except Exception, e:
        logger.error(unicode(e))
        sys.exit(1)





# -*- coding: utf-8 -*-

from __future__ import absolute_import
from collections import OrderedDict


import logging

logger = logging.getLogger(__name__)


class Service(object):

    def __init__(self, config=None, project_name=None):
        self.project_name = project_name
        self.config = config

    def get_template(self):
        return self._get_task_or_config(request_type="config")

    def get_task(self):
        return self._get_task_or_config(request_type="task")

    def _get_task_or_config(self, request_type="task"):
        templates = []
        for name, service in self.config.get('services', {}).items():
            new_service = self._create(request_type, name, service)
            if new_service:
                templates.append(new_service)
            if service.get('links'):
                templates += self._create_alias_templates(request_type, service['links'])
        return templates

    def _create_alias_templates(self, request_type, links):
        '''
        If a service defines aliased links, create a service template for each alias.

        :param request_type: will be 'config' or 'task'
        :param links: list of links
        :return: list of templates
        '''
        templates = []
        for link in links:
            if not ':' in link:
                continue
            service_name, alias = link.split(':')
            alias_config = self.config['services'].get(service_name)
            if alias_config:
                new_service = self._create(request_type, alias, alias_config)
                if new_service:
                    templates.append(new_service)
        return templates

    def _create(self, type, name, service):
        '''
        Create a Kubernetes service template or playbook task
        '''

        template = {}
        ports = self._get_ports(service)
        options = service.get('options', {}).get('kube', {})
        state = options.get('state', 'present')

        if ports:
            labels = dict(
                app=self.project_name,
                service=name
            )
            if type == 'config' and state != 'absent':
                template = dict(
                    apiVersion="v1",
                    kind="Service",
                    metadata=dict(
                        name=name,
                        labels=labels.copy()
                    ),
                    spec=dict(
                        selector=labels.copy(),
                        ports=ports,
                    )
                )

                for port in ports:
                    if port['port'] != port['targetPort']:
                        template['spec']['type'] = 'LoadBalancer'
            elif type == 'task':
                template = dict(
                    kube_service=OrderedDict(
                        service_name=name,
                        ports=ports,
                        selector=labels.copy()
                    )
                )

                if service.get('labels'):
                    template['kube_service']['labels'] = service.get('labels')

                load_balancer = False
                for port in ports:
                    if port['port'] != port['targetPort']:
                        load_balancer = True

                if load_balancer:
                    template['kube_service']['type'] = 'LoadBalancer'

        return template

    def _get_ports(self, service):
        # TODO - handle port ranges
        ports = []
        for port in service.get('ports', []):
            if isinstance(port, str) and ':' in port:
                parts = port.split(':')
                if not self._port_in_list(parts[0], ports):
                    ports.append(dict(port=int(parts[0]), targetPort=int(parts[1]), name='port-%s' % parts[0]))
            else:
                if not self._port_in_list(port, ports):
                    ports.append(dict(port=int(port), targetPort=int(port), name='port-%s' % port))

        for port in service.get('expose', []):
            if not self._port_in_list(port, ports):
                ports.append(dict(port=int(port), targetPort=int(port), name='port-%s' % port))
        return ports

    @staticmethod
    def _port_in_list(port, ports):
        found = False
        for p in ports:
            if p['port'] == int(port):
                found = True
                break
        return found
